/********************************************************************************
 * Copyright (c) Erik Kunze 1999
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *******************************************************************************/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifndef lint
static char rcsid[] = "$Id: chkrom.c,v 1.3 1999/03/07 16:42:26 erik Rel $";
#endif

/********************************************************************************
 * chkrom - Calculate checksum over a ROM image.
 *******************************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

/********************************************************************************
 * Defines
 *******************************************************************************/

#define ZX_Major				0
#define ZX_Minor				2

#define OPTIONS					"hvx"

/********************************************************************************
 * Forwards
 *******************************************************************************/

static int scanArgs(int, char **);
static void version(void);
static void copyright(void);
static void usage(void);
static int calc(void);
static int calcone(char *);

/********************************************************************************
 * Globals
 *******************************************************************************/

char *PrgName = NULL;					/* Basename of the program */

/********************************************************************************
 * Locals
 *******************************************************************************/

static struct option long_options[] = {
	{ "help",    no_argument,       0, (int)'h' },
	{ "version", no_argument,       0, (int)'v' },
	{ "hex",     no_argument,       0, (int)'x' },
};

static char **romFiles;
static int numFiles;
static int hex = 0;

/********************************************************************************
 * main() - Main entry point.
 *
 * Input:  argc - # of command line arguments
 *         argv - pointer to command line arguments
 *
 * Return: 0 - o.k.
 *         1 - error
 *******************************************************************************/

int
main(int argc, char ** argv)
{
	int ret = 1;

	/* Parse command line arguments */
	if (scanArgs(argc, argv) != -1)
	{
		/* Print out copyright message */
		copyright();

		/* Calculate checksum */
		ret = calc();
	}

	return ret;
}

/********************************************************************************
 * scanArgs() - Scan command line for arguments
 *
 * Input:  ac - # of command line arguments
 *         av - pointer to command line arguments
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

static int
scanArgs(int ac, char *av[])
{
	int opt;
	int idx = 0;

	if (!(PrgName = strrchr(av[0], '/')))
	{
		PrgName = av[0];
	}
	else
	{
		PrgName++;
	}

	opterr = 0;
	optind = 0;
	while ((opt = getopt_long(ac, av, OPTIONS, long_options, &idx)) != EOF)
	{
		switch (opt)
		{
			case 'h':
				usage();
				return -1;
			case 'v':
				version();
				return -1;
			case 'x':
				hex++;
				break;
			default:
				usage();
				return -1;
		}
	}

#ifdef DEBUG
	if (optind < ac)
	{
		int i = optind;

		(void)fprintf(stderr, "[scanArgs] additional arguments: ");
		while (i < ac)
		{
			(void)fprintf(stderr, "%s ", av[i++]);
		}
		(void)fputc('\n', stderr);
	}
#endif /* DEBUG */

	if (optind < ac)
	{
		romFiles = &av[optind];
		numFiles = ac - optind;
		return 0;
	}

	usage();
	return -1;
}

/********************************************************************************
 * version() -  Print out the version information.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
version(void)
{
	copyright();
	(void)fprintf(stderr, "This is ChkRom %d.%d.\n", ZX_Major, ZX_Minor);
}

/********************************************************************************
 * copyright() - Print out a copyright message.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
copyright(void)
{
	(void)fprintf(stderr,
	"ChkRom is copyright (c) 1999 Erik Kunze\n"
	"ChkRom comes with no warranty; see the file COPYRIGHT for details.\n\n");
}

/********************************************************************************
 * usage() - Print out a usage message.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
usage(void)
{
	(void)fprintf(stderr, "Usage: %s [options] files\n",
				  PrgName);
	(void)fprintf(stderr, "Available options:\n"
	"  -h, --help       display this message\n"
	"  -v, --version    display version number and expiration date\n"
	"  -x, --hex        print checksum in hex mode\n"
	);
}

/********************************************************************************
 * calc() - Call for each file calcone().
 *
 * Input:  -
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

static int
calc(void)
{
	int ret = 0, i;

	for (i = 0; i < numFiles; i++)
	{
		ret |= calcone(romFiles[i]);
	}

	return ret;
}

/********************************************************************************
 * calcone() - Read rom file, calculate and print checksum.
 *
 * Input:  fname - name of ROM file
 *
 * Return:  0  - o.k.
 *         <>0 - error
 *******************************************************************************/

static int
calcone(char *fname)
{
	FILE *fp;
	int checksum, tmp, c, length;

	/* Try to open input file. */
	if (!(fp = fopen(fname, "rb")))
	{
		(void)fprintf(stderr, "%s: couldn't open <%s> for reading",
					  PrgName, fname);
		perror("  --> reason");
		return 1;
	}

	checksum = 0;
	length = 0;

	while ((c = fgetc(fp)) != EOF)
	{
		length++;
		tmp = checksum & 0x01;
		checksum >>= 1;
		if (tmp)
		{
			checksum |= 0x8000;
		}
		checksum ^= c;
	}

	(void)printf(hex ? "%s: %04X, %6d bytes\n"
					 : "%s: %5d, %6d bytes\n",
				 fname, checksum & 0xffff, length);

	(void)fclose(fp);
	return 0;
}

/* EOF *************************************************************************/
