/* sna.c: routines for dealing with .sna format files
   Copyright (c) 1997-8 Philip Kendall

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: pak21@cam.ac.uk (until at least June 1999)
   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England

*/

#include <stdio.h>
#include <string.h>

#include "snapconv.h"
#include "snapinfo.h"

#include "sna.h"

/* Process the 27 byte header on both 48K and 128K .sna files */

int sna_readHeader(FILE *f,struct snapinfo *sptr)
{
  unsigned char Header[sna_headerSize];

  if(fread(Header,1,sna_headerSize,f)!=sna_headerSize) return ERROR_READ;

  sptr->i=Header[0];
  sptr->hl_=Header[1]+Header[2]*0x100;
  sptr->de_=Header[3]+Header[4]*0x100;
  sptr->bc_=Header[5]+Header[6]*0x100;
  sptr->f_=Header[7]; sptr->a_=Header[8];
  sptr->hl=Header[ 9]+Header[10]*0x100;
  sptr->de=Header[11]+Header[12]*0x100;
  sptr->bc=Header[13]+Header[14]*0x100;
  sptr->iy=Header[15]+Header[16]*0x100;
  sptr->ix=Header[17]+Header[18]*0x100;

  /* Thanks to Paul Hodgson for pointing out the mistake here */
  sptr->iff1=sptr->iff2=(Header[19]&0x04)>>2;

  sptr->r=Header[20];
  sptr->f=Header[21]; sptr->a=Header[22];
  sptr->sp=Header[23]+Header[24]*0x100;
  sptr->im=Header[25]; sptr->outULA=Header[26]&0x07;

  if(verbose>=2)
    {
      fprintf(stderr," I =0x%02x\n",sptr->i);
      fprintf(stderr,"HL'=0x%04x DE'=0x%04x\n",sptr->hl_,sptr->de_);
      fprintf(stderr,"BC'=0x%04x AF'=0x%02x%02x\n",
	      sptr->bc_,sptr->a_,sptr->f_);
      fprintf(stderr,"HL =0x%04x DE =0x%04x\n",sptr->hl,sptr->de);
      fprintf(stderr,"BC =0x%04x\n",sptr->bc);
      fprintf(stderr,"IX =0x%04x IY =0x%04x\n",sptr->ix,sptr->iy);
      fprintf(stderr,"IFF2=%d      R =0x%02x\n",sptr->iff2,sptr->r);
      fprintf(stderr,"AF =0x%02x%02x SP =0x%04x (Incorrect for 48K snaps)\n",
	      sptr->a,sptr->f,sptr->sp);
      fprintf(stderr,"IM %d       Border %d\n",sptr->im,sptr->outULA);
    }

  return ERROR_NONE;

}

/* Remember to decrease SP if calling this to deal with 48K .sna files */

int sna_writeHeader(FILE *f,struct snapinfo *sptr)
{
 fputc(sptr->i,f);

 writeInt(sptr->hl_,f); writeInt(sptr->de_,f);
 writeInt(sptr->bc_,f); fputc(sptr->f_,f); fputc(sptr->a_,f);

 writeInt(sptr->hl,f); writeInt(sptr->de,f);
 writeInt(sptr->bc,f); writeInt(sptr->iy,f);
 writeInt(sptr->ix,f);

 fputc( (sptr->iff2)<<2 ,f);
 fputc(sptr->r,f); fputc(sptr->f,f); fputc(sptr->a,f);

 writeInt(sptr->sp,f);

 fputc(sptr->im,f); fputc( (sptr->outULA)&0x07 , f);

 return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int sna_read128Header(FILE *f,struct snapinfo *sptr)
{
  unsigned char Header[sna_128HeaderSize];

  if(fread(Header,1,sna_128HeaderSize,f)!=sna_128HeaderSize) return ERROR_READ;

  sptr->pc=Header[0]+Header[1]*0x100;
  sptr->out7ffd=Header[2];

  if(verbose>=2)
    fprintf(stderr,"PC =0x%04x\nLast out to 0x7ffd=0x%02x\n",
	    sptr->pc,sptr->out7ffd);

  return ERROR_NONE;

}

int sna_write128Header(FILE *f,struct snapinfo *sptr)
{
  writeInt(sptr->pc,f);
  fputc(sptr->out7ffd,f); fputc('\0',f);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE ;

}

int sna_read48Memory(FILE *f,struct snapinfo *sptr,unsigned char *buffer)
{
  int offset=(sptr->sp)-0x4000;

  if(fread(buffer,1,0xc000,f)!=0xc000) return ERROR_READ;

  sptr->pc=buffer[offset]+buffer[offset+1]*0x100;

  /* replace corrupted bytes with zero -- apparently this can 'rescue' some
     snaps which were saved at critical points
  */

  buffer[offset]=buffer[offset+1]=0;

  /* and increase SP to account for PC being unstacked */

  (sptr->sp)+=2;

  if(verbose>=2)
    fprintf(stderr,"PC=0x%04x\nCorrect SP=0x%04x\n",sptr->pc,sptr->sp);

  return ERROR_NONE;

}

int sna_write48Memory(FILE *f,struct snapinfo *sptr,unsigned char *buffer)
{
  unsigned int offset=(sptr->sp)-0x4000;

  /* Write the bytes before where PC will be stacked */

  (void)fwrite(buffer,offset,1,f);

  /* Write out PC */

  writeInt(sptr->pc,f);

  /* And write the rest */

  (void)fwrite(&buffer[offset+2],0xc000-offset-2,1,f);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}
