/* snapconv.c: Spectrum snapshot convertor v0.10
   Copyright (c) 1997-9 Philip Kendall

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: pak21@cam.ac.uk (until at least June 1999)
   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England

*/

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#include "hardware.h"
#include "snapconv.h"
#include "snapinfo.h"

#include "plusd.h"
#include "scr.h"
#include "sna.h"
#include "snp.h"
#include "sp.h"
#include "z80.h"
#include "zxs.h"

/* Disable DJGPP's argument globbing and environment loading */
#ifdef __DJGPP__
char** __crt0_glob_function(char* arg) {return 0;}
void __crt0_load_environment_file(char* name) {return;}
#endif                  /* #ifdef __DJGPP__ */

int getType(char *string);
int findString(char *toFind,const char **strings);

int z80_read(FILE *f,char *inName,struct snapinfo *sptr,unsigned char **pages,
	     romData *romPages);
int z80_write(FILE *f,char *outName,int version,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages);

int sna_read(FILE *f,char *inName,struct snapinfo *sptr,unsigned char **pages,
	     romData *romPages);
int sna_write(FILE *f,char *outName,int version,struct snapinfo *sptr,
               unsigned char **pages,romData *romPages);
int sna48_read(FILE *f,char *inName,struct snapinfo *sptr,
              unsigned char **pages,romData *romPages);
int sna48_write(FILE *f,char *outName,struct snapinfo *sptr,
               unsigned char **pages,romData *romPages);
int sna128_read(FILE *f,char *inName,struct snapinfo *sptr,
              unsigned char **pages,romData *romPages);
int sna128_write(FILE *f,char *outName,struct snapinfo *sptr,
               unsigned char **pages,romData *romPages);

int snp_read2(FILE *f,char *inName,struct snapinfo *sptr,
              unsigned char **pages,romData *romPages);
int snp_write2(FILE *f,char *outName,int version,struct snapinfo *sptr,
               unsigned char **pages,romData *romPages);

int sp_read2(FILE *f,char *inName,struct snapinfo *sptr,
	     unsigned char **pages,romData *romPages);
int sp_write2(FILE *f,char *outName,int version,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages);

int zxs_read(FILE *f,char *inName,struct snapinfo *sptr,
	     unsigned char **pages,romData *romPages);
int zxs_write(FILE *f,char *outName,int version,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages);

int plusd_read2(FILE *f,char *inName,struct snapinfo *sptr,
		unsigned char **pages,romData *romPages);
int plusd_write2(FILE *f,char *outName,int version,struct snapinfo *sptr,
		 unsigned char **pages,romData *romPages);

int construct48KBuffer(unsigned char **pages,unsigned char *buffer);
int construct48KPages(unsigned char *buffer, unsigned char **pages);
int printError(char *string,int errorNum);

/* the file formats supported by this program */

typedef int (*snapReadFunction )(FILE *f,char *inName,struct snapinfo *sptr,
				 unsigned char **pages,
				 romData *romPages);
typedef int (*snapWriteFunction)(FILE *f,char *outName,int version,
				 struct snapinfo *sptr,unsigned char **pages,
				 romData *romPages);

typedef struct formatInfo {
  const char *extension;     /* Extension used to identify this format
			        0 => no extension, must select with -i,-o */
  const char *ID;		     /* String used to identify format with -i,-o */
  snapReadFunction readf;    /* Function to read this format */
  snapWriteFunction writef;  /* Function to write this format */
} formatInfo;

/* A reference number for each of the supported types */

enum {

  TYPE_Z80   =0,
  TYPE_SNA   =1,
  TYPE_SNA48 =2,
  TYPE_SNA128=3,
  TYPE_SNP   =4,
  TYPE_SP    =5,
  TYPE_ZXS   =6,
  TYPE_PLUSD =7

};

/* Make sure the order here matches that in the enum! */

formatInfo supportedTypes[] =
{
  { "z80", "z80",    z80_read,    z80_write    },
  { "sna", "sna",    sna_read,    sna_write    },
  { 0    , "sna48",  sna48_read , sna_write    },
  { 0    , "sna128", sna128_read, sna_write    },
  { "snp", "snp",    snp_read2,   snp_write2   },
  { "sp" , "sp",     sp_read2,    sp_write2    },
  { "zxs", "zxs",    zxs_read,    zxs_write    },
  { 0    , "plusd",  plusd_read2, plusd_write2 },
  { "scr", "scr",    0,           scr_write    }
};

const int numSupportedTypes=9;

/* There are 3 entries for .sna in the above table to allow for the
   possibility that input is being read from a device, when the
   file size will not be available, so auto-detection of 48K/128K
   is not possible; note that all three entries use sna_write,
   so the correct type of snap will be written even if the output
   format is set to `sna128' and it's a 48K snap, or vice-versa */

/* The machines it is possible to select with -m */

const char *supportedMachines[] = { "48", "128", "plus2", "plus2a", "plus3", 0 };

/* The list of peripherals it is possible to select with -p */

const char *supportedPeriph[] = { "if1", "mgt", "samram", 0 };

/* The possible error messages; make sure the order in here matches
   the enum defined in snapconv.h */

const char *errorMessages[] = {
  "No error",
  "Invalid command line option",
  "Input and output filenames must be specified",
  "Unspecified input format",
  "Unspecified output format",
  "Unknown input format",
  "Unknown output format",
  "Invalid output version",
  "Couldn't open for input",
  "Couldn't open for output",
  "Out of memory",
  "Unknown page number",
  "Duplicate pages",
  "Unknown length for .sna file",
  "Read error",
  "Write error",
  "Corrupt compressed data",
  "Internal error: invalid special hardware map",
  "Unknown header length",
  "Unknown hardware specified",
  "Subchunk lengths do not add up to specified length",
  "Memory page is of incorrect length",
  "Sorry: can't handle this compression type",
  "Internal error: attempt to write unknown chunk type",
  "Internal error: no form type of parent chunk",
  "Error from zlib compression routines",
  "Sorry: can't read this file format"
};

const int numErrorMessages=27;

/* The length and Adler-32 checksum of each of the ROMs; make sure the
   order in here matches the order in the enum defined in snapconv.h */

romInfoStruct romInfo[] =
{

 /* length   checksum */

  { 0x4000, 0x5e9d65d3 }, /* 48K ROM */
  { 0x4000, 0x88eb7353 }, /* 128K ROM 0 */
  { 0x4000, 0xb88d87b3 }, /* 128K ROM 1 */
  { 0x4000, 0x1e9c4127 }, /* +2 ROM 0 */
  { 0x4000, 0x003982f7 }, /* +2 ROM 1 */
  { 0x4000, 0x808bec4b }, /* +2A/+3 ROM 0 */
  { 0x4000, 0x8d52e15a }, /* +2A/+3 ROM 1 */
  { 0x4000, 0x9172aa1e }, /* +2A/+3 ROM 2 */
  { 0x4000, 0x76984178 }, /* +2A/+3 ROM 3 */
  { 0x2000, 0xce114090 }, /* Interface 1 ROM */
  { 0x4000, 0x044b6458 }, /* Disciple ROM */
  { 0x4000, 0x845c7289 }, /* +D ROM */
  { 0x4000, 0x755a7f51 }, /* SamRam ROM 0 */
  { 0x4000, 0x55e8c355 }, /* SamRam ROM 1 */
  { 0x2000, 0x4a3ba3bb }  /* Multiface 128 ROM */

};

int verbose; /* How many messages to produce */

int main(int argc,char **argv)
{
  int i,error=0,outVersion=0,inType=-1,outType=-1,rCorrection=0,machine=-1,
      periphOn=0,periphOff=0,temp;
  FILE *f; struct snapinfo snap;
  unsigned char *pages[NUMRAMPAGES]; romData romPages[NUMROMPAGES];
  char *inName,*outName;

  verbose=0;

  opterr=0;		/* supress error messages from getopt */

  while( (i=getopt(argc,argv,"123i:m:o:p:q:r:v")) != -1 )
    {
      if(verbose>=2) fprintf(stderr,"Parsing option `-%c'\n",i);
      switch(i)
	{
	  case '1': outVersion=1; break;
	  case '2': outVersion=2; break;
	  case '3': outVersion=3; break;
	  case 'i': inType=getType(optarg);
   	            if(inType==-1) return printError(0,ERROR_COMMANDLINE);
	            break;
	  case 'm': machine=findString(optarg,supportedMachines);
	            if(machine==-1) return printError(0,ERROR_COMMANDLINE);
		    break;
	  case 'o': outType=getType(optarg);
	            if(outType==-1) return printError(0,ERROR_COMMANDLINE);
	            break;
	  case 'p': temp=findString(optarg,supportedPeriph);
	            if(temp==-1) return printError(0,ERROR_COMMANDLINE);
		    periphOn |= (1<<temp);
		    break;
	  case 'q': temp=findString(optarg,supportedPeriph);
	            if(temp==-1) return printError(0,ERROR_COMMANDLINE);
		    periphOff |= (1<<temp);
		    break;
	  case 'r': rCorrection=atoi(optarg); break;
	  case 'v': verbose++;
	            fprintf(stderr,"Verbose level: %d\n",verbose);
	            break;
	  case '?': return printError(0,ERROR_COMMANDLINE);
	}
    }

  if(argc<optind+2) return printError(0,ERROR_FILENAMES);

  inName=argv[optind]; outName=argv[optind+1];

  if(verbose>=2)
    fprintf(stderr,"Input filename: `%s'\nOutput filename: `%s'\n",
	    inName,outName
	    );

  /* If not specified on the command line,
     find out what sort of file we want to convert from/to */

  if(inType==-1)
    {
      char *inExt=strrchr(inName,'.');
      if(!inExt) return printError(inName,ERROR_NOINFORMAT);

      for(i=0;i<numSupportedTypes;i++)
	if(supportedTypes[i].extension &&
	   !strcasecmp(&inExt[1],supportedTypes[i].extension) ) inType=i;
    }

  if(outType==-1)
    {
      char *outExt=strrchr(outName,'.');
      if(!outExt) return printError(outName,ERROR_NOOUTFORMAT);

      for(i=0;i<numSupportedTypes;i++)
	if(supportedTypes[i].extension &&
	   !strcasecmp(&outExt[1],supportedTypes[i].extension) ) outType=i;
    }

  if(inType==-1)  return printError(inName ,ERROR_UNKNOWNINFORMAT );
  if(outType==-1) return printError(outName,ERROR_UNKNOWNOUTFORMAT);

  if( outType==TYPE_Z80 && outVersion==0) outVersion=Z80_DEFAULTOUTVERSION;

  if( outType==TYPE_SP )
    {
      switch(outVersion)
	{
	  case 0 : outVersion=SP_DEFAULTOUTVERSION; break;
	  case 1 :
	  case 2 : break;
	  default: return printError(outName,ERROR_OUTVERSION);
	}
    }

  /* Set values to defaults */
  for(i=0;i<NUMRAMPAGES;i++) pages[i]=NULL; /* Page not read in */
  for(i=0;i<NUMROMPAGES;i++) { romPages[i].present=0; romPages[i].ptr=NULL; }
  snapinfoDefault(&snap);    /* Sensible values */

  if((f=fopen(inName,"rb"))==NULL) return printError(inName,ERROR_INOPEN);
  if(verbose>=1) fprintf(stderr,"File `%s' opened for input\n",inName);

  /* Read in the selected snap format */
  if(supportedTypes[inType].readf)
    {
      error=(supportedTypes[inType].readf)(f,inName,&snap,pages,romPages);
    }
  else
    return printError(inName,ERROR_CANTREAD);

  if(error!=ERROR_NONE)
    {
      for(i=0;i<NUMRAMPAGES;i++) if(pages[i]) free(pages[i]);
      for(i=0;i<NUMROMPAGES;i++) if(romPages[i].ptr) free(romPages[i].ptr);
      return printError(inName,error);
    }

  if(verbose>=1) fprintf(stderr,"File `%s' read successfully\n",inName);

  /* Change to the requested machine */
  if(machine!=-1) snap.machine=machine;

  /* Turn any peripherals on and off */
  snap.periph|=periphOn;
  snap.periph&=~periphOff;

  /* Apply any requested corrections to R */
  snap.r+=rCorrection;

  if(freopen(outName,"wb",f)==NULL)
    {
      for(i=0;i<NUMRAMPAGES;i++) if(pages[i]) free(pages[i]);
      for(i=0;i<NUMROMPAGES;i++) if(romPages[i].ptr) free(romPages[i].ptr);
      return printError(outName,ERROR_OUTOPEN);
    }
  if(verbose>=1) fprintf(stderr,"File `%s' opened for output\n",outName);

  /* Write out the selected snap format */
  error=(supportedTypes[outType].writef)
          (f,outName,outVersion,&snap,pages,romPages);

  if(error!=ERROR_NONE)
    {
      for(i=0;i<NUMRAMPAGES;i++) if(pages[i]) free(pages[i]);
      for(i=0;i<NUMROMPAGES;i++) if(romPages[i].ptr) free(romPages[i].ptr);
      return printError(outName,error);
    }

  if(verbose>=1) fprintf(stderr,"File `%s' written successfully\n",outName);

  for(i=0;i<NUMRAMPAGES;i++) if(pages[i]) free(pages[i]);
  for(i=0;i<NUMROMPAGES;i++) if(romPages[i].ptr) free(romPages[i].ptr);

  fclose(f);

  return ERROR_NONE;

}

int getType(char *string)
{
  int i;

  for(i=0;i<numSupportedTypes;i++)
    {
      if(!strcasecmp(string,supportedTypes[i].ID)) return i;
    }
  return -1;
}

int findString(char *toFind,const char **strings)
{
  const char **ptr=strings;

  while(*ptr!=NULL)
    {
      if(!strcasecmp(toFind,*ptr)) return (ptr-strings);
      ptr++;
    }
  return -1;
}

/* Overall .z80 control functions */

int z80_read(FILE *f,char *inName,struct snapinfo *sptr,
             unsigned char **pages,romData *romPages)
{

  int error,page; unsigned length; unsigned char *buffer;

  if((error=z80_readHeader(f,sptr))!=0) return error;
  if(verbose>=1) fprintf(stderr,".z80 version %d header read sucessfully\n",
			 sptr->version);

  if(sptr->version>=2)
    {while(!feof(f) && !ferror(f) && !error)
      {

	error=z80_readBlock(f,&page,&buffer,&length);

	if(feof(f)) break;
	if(error!=ERROR_NONE) return error;

	if(verbose>=1) fprintf(stderr,"Read .z80 block %d\n",page);

	error=z80_parsePage(page,buffer,length,sptr,pages,romPages);
	if(error!=ERROR_NONE)
	  {
	    free(buffer);
	    return error;
	  }

      }

    }
  else			/* v1 .z80 file => 1 block only */
    {if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
      return ERROR_MEMORY;

    error=z80_readv1Data(f,buffer,sptr->compressed);
    if(error!=ERROR_NONE)
      {
	free(buffer);
	return error;
      }
    if(verbose>=1) fprintf(stderr,".z80 memory dump read successfully\n");

    error=construct48KPages(buffer,pages);
    if(error!=ERROR_NONE)
      {
	free(buffer);
	return error;
      }

    free(buffer);

    }

  return ERROR_NONE;

}

int z80_write(FILE *f,char *outName,int version,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages)
{

  int i,error,page;

  switch(version)
    {
      case 1:
	error=unsupportedHardware(sptr,z80_version1Map,"`.z80' v1");
	if(error!=ERROR_NONE) return error;
	break;
      case 2:
	error=unsupportedHardware(sptr,z80_version2Map,"`.z80' v2");
	if(error!=ERROR_NONE) return error;
	break;
      case 3:
	error=unsupportedHardware(sptr,z80_version3Map,"`.z80' v3");
	if(error!=ERROR_NONE) return error;
	break;
    }

  switch(version)
    {
      case 1: error=z80_writev1Header(f,sptr,0); break;
      case 2: error=z80_writev2Header(f,sptr,0); break;
      case 3: error=z80_writev3Header(f,sptr  ); break;
    }

  if(verbose>=1)
    fprintf(stderr,".z80 version %d header written successfully\n",
	    version);

  if(version>=2)
    {

      for(i=0;i<NUMROMPAGES;i++) if(romPages[i].ptr)
	{
	  page=-1;

	  switch(i)
	    {
	      case ROM_48K: if(sptr->machine==MACHINE_48K) page=0; break;
	      case ROM_128K_0: if(sptr->machine!=MACHINE_48K) page=2; break;
	      case ROM_128K_1: if(sptr->machine!=MACHINE_48K) page=0; break;
	      case ROM_IF1: if(sptr->periph&PERIPH_IF1) page=1; break;
	      case ROM_DISCIPLE:
		if(sptr->periph&PERIPH_MGT && sptr->mgttype!=16) page=1; break;
	      case ROM_PLUSD:
		if(sptr->periph&PERIPH_MGT && sptr->mgttype==16) page=1; break;
	      case ROM_SAMRAM_0: if(sptr->periph&PERIPH_SAMRAM) page=2; break;
	      case ROM_SAMRAM_1: if(sptr->periph&PERIPH_SAMRAM) page=3; break;
	      case ROM_MULTIFACE: page=11; break;
	    }

	  if(page!=-1)
	    {
	      error=z80_writeBlock(f,page,romPages[i].ptr,romInfo[i].length);
	      if(verbose>=1)
		fprintf(stderr,".z80 block number %d written successfully\n",
			page);
	    }
	}

      for(i=0;i<NUMRAMPAGES;i++) if(pages[i])
	{

	  page=i;

	  /* Re-map 48K pages to the way they exist in a 48K .z80 file
	     (page 1=8000-bfff, page 2=c000-ffff, page 5=4000-7fff)
	     I use 0=c000-ffff, 2=8000-bfff, 5=4000-7fff, as this
	     more closely corresponds to the 128K setup. I also re-map
	     Gerton's page 0 to my page 1, just in case...
	     */

	  if(sptr->machine==MACHINE_48K)
	    {
	      switch(page)
		{
		  case 0: page=2; break;
		  case 2: page=1; break;

		  case 1: page=0; break;
		}
	    }

	  error=z80_writeBlock(f,page+3,pages[i],0x4000);
	  if(error!=ERROR_NONE) return error;
	  if(verbose>=1)
	    fprintf(stderr,".z80 block number %d written successfully\n",
		    page+3
		   );

	}
    }
  else			/* version=1 */
    {
      error=z80_writev1Data(f,pages[5],pages[2],pages[0]);
      if(error!=ERROR_NONE) return error;
      if(verbose>=1) fprintf(stderr,".z80 memory dump written successfully\n");
    }

  return ERROR_NONE;

}

/* Overall .sna control functions */

int sna_read(FILE *f,char *inName,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages)
{struct stat fileInfo; long length;

 stat(inName,&fileInfo);
 length=fileInfo.st_size;
 if(verbose>=1) fprintf(stderr,"Length of .sna file: %ld\n",length);

 switch(length)
   {
     case  49179: if(verbose>=2) fprintf(stderr,"Reading 48K .sna file\n");
                  return sna48_read(f,inName,sptr,pages,romPages);
     case 131103:
     case 147487: if(verbose>=2) fprintf(stderr,"Reading 128K .sna file\n");
                  return sna128_read(f,inName,sptr,pages,romPages);
         default: return ERROR_SNALENGTH;
   }
}

int sna_write(FILE *f,char *outName,int version,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages)
{
  int error;

  error=unsupportedHardware(sptr,UNSUPPORTED_128KNOPERIPHS,"`.sna'");
  if(error!=ERROR_NONE) return error;

  if(sptr->machine==MACHINE_128K)
    {
      if(verbose>=1) fprintf(stderr,"Writing 128K .sna file\n");
      return sna128_write(f,outName,sptr,pages,romPages);
    }
  else
    {
      if(verbose>=1) fprintf(stderr,"Writing 48K .sna file\n");
      return sna48_write(f,outName,sptr,pages,romPages);
    }
}

/* 48K sna functions */

int sna48_read(FILE *f,char *inName,struct snapinfo *sptr,
              unsigned char **pages,romData *romPages)
{int error; unsigned char *buffer;

 if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
   return ERROR_MEMORY;

 error=sna_readHeader(f,sptr);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,".sna header read successfully\n");

 error=sna_read48Memory(f,sptr,buffer);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,".sna memory dump read successfully\n");

 error=construct48KPages(buffer,pages);
 if(error!=ERROR_NONE) { free(buffer); return error; }

 free(buffer);

 return ERROR_NONE;

}

int sna48_write(FILE *f,char *outName,struct snapinfo *sptr,
		unsigned char **pages,romData *romPages)
{unsigned char *buffer; int error;

 if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
   return ERROR_MEMORY;

 construct48KBuffer(pages,buffer);

 /* Decrease SP to deal with PC being stacked */

 (sptr->sp)-=2;

 error=sna_writeHeader(f,sptr);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,".sna header written successfully\n");

 error=sna_write48Memory(f,sptr,buffer);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,".sna memory dump written successfully\n");

 /* Restore SP */
 (sptr->sp)+=2;

 free(buffer);

 return ERROR_NONE;

}

/* 128K sna functions */

int sna128_read(FILE *f,char *inName,struct snapinfo *sptr,
                unsigned char **pages,romData *romPages)
{int error,page,i,j; unsigned char *buffer;

 if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
   return ERROR_MEMORY;

 error=sna_readHeader(f,sptr);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,"48K .sna header read successfully\n");

 /* This is a 128K snap */
 sptr->machine=MACHINE_128K;

 /* No special processing is needed for the main memory dump in
    128K sna files */
 if(fread(buffer,0xc000,1,f)!=1) {free(buffer); return ERROR_READ;}
 if(verbose>=1) fprintf(stderr,"48K .sna memory dump read successfully\n");

 error=sna_read128Header(f,sptr);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,"128K .sna header read successfully\n");

 /* Decode the last byte output to port #7FFD, so I know which
    pages are to follow */
 page= (sptr->out7ffd)&0x07;

 pages[5]=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
 if(!pages[5]) { free(buffer); return ERROR_MEMORY; }

 pages[2]=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
 if(!pages[2]) { free(buffer); free(pages[5]); return ERROR_MEMORY; }

 memcpy(pages[5],&buffer[0x0000],0x4000);
 memcpy(pages[2],&buffer[0x4000],0x4000);

 if(page==5 || page==2)
   {
     if( memcmp(pages[page],&buffer[0x8000],0x4000) )
       return ERROR_DUPLICATEPAGE;
   }
 else
   {
     pages[page]=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
     if(!pages[page])
       {
	 free(buffer); free(pages[2]); free(pages[5]);
	 return ERROR_MEMORY;
       }
     memcpy(pages[page],&buffer[0x8000],0x4000);
   }
 if(verbose>=1) fprintf(stderr,"Page number %d read successfully\n",page);

 for(i=0;i<=7;i++)
   {
     if(i==2 || i==5 || i==page) continue; /* Already got this page */

     pages[i]=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
     if(!pages[i])
       {
	 for(j=0;j<7;j++) if(pages[i]) free(pages[i]); free(buffer);
	 return ERROR_MEMORY;
       }
     if( fread(pages[i],0x4000,1,f)!=1 )
       {
	 for(j=0;j<7;j++) if(pages[i]) free(pages[i]); free(buffer);
	 return ERROR_READ;
       }
     if(verbose>=1) fprintf(stderr,"Page number %d read successfully\n",i);
   }

 free(buffer);

 return ERROR_NONE;

}

int sna128_write(FILE *f,char *outName,struct snapinfo *sptr,
		 unsigned char **pages,romData *romPages)
{int i,error,page; unsigned char *buffer;

 if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
   return ERROR_MEMORY;

 /* Which memory page is currently selected? */

 page=(sptr->out7ffd)&0x07;

 if(pages[   5]) memcpy(&buffer[0x0000],pages[   5],0x4000);
   else memset(&buffer[0x0000],0,0x4000);
 if(pages[   2]) memcpy(&buffer[0x4000],pages[   2],0x4000);
   else memset(&buffer[0x4000],0,0x4000);
 if(pages[page]) memcpy(&buffer[0x8000],pages[page],0x4000);
   else memset(&buffer[0x8000],0,0x4000);

 error=sna_writeHeader(f,sptr);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,"48K .sna header written successfully\n");

 /* Dump the paged in memory */

 if( fwrite(buffer,0xc000,1,f)!=1) { free(buffer); return ERROR_WRITE; }
 if(verbose>=1) fprintf(stderr,"48K .sna memory dump written successfully\n");

 /* And now the 128K header */

 error=sna_write128Header(f,sptr);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,"128K .sna header written successfully\n");

 /* Write out the pages not already in the file */

 memset(buffer,0,0x4000); /* A blank page; used if page n not read in */

 for(i=0;i<=7;i++)
   {
     if(i==2 || i==5 || i==page) continue; /* Already done this page */
     if(pages[i]) (void)fwrite(pages[i],0x4000,1,f);
     else (void)fwrite(buffer,0x4000,1,f);
     if(verbose>=1) fprintf(stderr,"Page number %d written successfully\n",i);
   }

 if(ferror(f)) { free(buffer); return ERROR_WRITE; }

 free(buffer);

 return ERROR_NONE;

}

/* Overall .snp control functions */

int snp_read2(FILE *f,char *inName,struct snapinfo *sptr,
              unsigned char **pages,romData *romPages)
{
  int error; unsigned char *buffer;

  if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
    return ERROR_MEMORY;

  error=snp_read(f,sptr,buffer);
  if(error!=ERROR_NONE) { free(buffer); return error; }
  if(verbose>=1) fprintf(stderr,".snp file read successfully\n");

  error=construct48KPages(buffer,pages);
  if(error!=ERROR_NONE) { free(buffer); return error; }

  free(buffer);

  return ERROR_NONE;

}

int snp_write2(FILE *f,char *outName,int version,struct snapinfo *sptr,
               unsigned char **pages,romData *romPages)
{
  unsigned char *buffer; int error;

  error=unsupportedHardware(sptr,UNSUPPORTED_48KNOPERIPHS,"`.snp'");
  if(error!=ERROR_NONE) return error;

  if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
    return ERROR_MEMORY;

  construct48KBuffer(pages,buffer);

  error=snp_write(f,sptr,buffer);
  if(error!=ERROR_NONE) { free(buffer); return error; }
  if(verbose>=1) fprintf(stderr,".snp file written successfully\n");

  free(buffer);

  return ERROR_NONE;

}

/* Overall .sp control functions */

int sp_read2(FILE *f,char *inName,struct snapinfo *sptr,
             unsigned char **pages,romData *romPages)
{int error; unsigned char *buffer;

 if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
   return ERROR_MEMORY;

 error=sp_read(f,sptr,buffer);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,".sp file read successfully\n");

 error=construct48KPages(buffer,pages);
 if(error!=ERROR_NONE) { free(buffer); return error; }

 free(buffer);

 return ERROR_NONE;

}

int sp_write2(FILE *f,char *outName,int version,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages)
{
  unsigned char *buffer; int error;

  error=unsupportedHardware(sptr,UNSUPPORTED_48KNOPERIPHS,"`.sp'");
  if(error!=ERROR_NONE) return error;

  if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
    return ERROR_MEMORY;

  construct48KBuffer(pages,buffer);

  error=sp_write(f,sptr,version,buffer);
  if(error!=ERROR_NONE) { free(buffer); return error; }
  if(verbose>=1) fprintf(stderr,".sp file written successfully\n");

  free(buffer);

  return ERROR_NONE;

}

/* .zxs control functions */

int zxs_read(FILE *f,char *inName,struct snapinfo *sptr,
	     unsigned char **pages,romData *romPages)
{
  unsigned long rd=0; zxs_chunkList *p; int error,majorVersion,minorVersion;

  error=zxs_readChunk(f,sptr,pages,&rd);
  if(error!=ERROR_NONE) return error;

  majorVersion=(sptr->version)/0x100;
  minorVersion=(sptr->version)%0x100;

  if(majorVersion>zxs_majorVersion)
    {
      printf("Warning: major version of %s (%d) is greater than that\n"
	     "         supported by this program (%d). Conversion may not work\n",
	     inName,majorVersion,zxs_majorVersion);
    }
  else if(majorVersion==zxs_majorVersion && minorVersion>zxs_minorVersion)
    {
      printf("Warning: minor version of %s (%d) is greater than that\n"
	     "         supported by this program (%d). Conversion may not work\n",
	     inName,minorVersion,zxs_minorVersion);
    }

  if(zxs_status&0x01)	/* Some unknown chunks found */
    {
      p=zxs_unknownChunks;
      printf("Unknown chunk(s) were found in %s; chunk ID(s):\n",inName);
      do
	{
	  printf("  %s\n",p->ID);
	}
      while( (p=p->next) != 0 );
      printf("\n");
    }

  if(zxs_status&0x02)	/* Some short chunks found */
    {
      p=zxs_shortChunks;
      printf("Chunk(s) in %s were shorter than expected and skipped; chunk ID(s):\n",
	     inName);
      do
	{
	  printf("  %s\n",p->ID);
	}
      while( (p=p->next) != 0 );
      printf("\n");
    }

  if(zxs_status&0x04)	/* Some long chunks found */
    {
      p=zxs_unknownChunks;
      printf("Chunk(s) in %s were longer than expected and some data ignored; chunk ID(s):\n",
	     inName);
      do
	{
	  printf("  %s\n",p->ID);
	}
      while( (p=p->next) != 0 );
      printf("\n");
    }

  zxs_cleanUp();

  return ERROR_NONE;

}

int zxs_write(FILE *f,char *outName,int version,struct snapinfo *sptr,
	      unsigned char **pages,romData *romPages)
{
  unsigned long length=0; zxs_chunkList *head=0,**sub; int error;

  error=unsupportedHardware(sptr,zxs_hardwareMap,"`.zxs'");
  if(error!=ERROR_NONE) return error;

  error=zxs_addChunkToList(&head,"RIFF");
  if(error!=ERROR_NONE) return error;

  sub=&(head->sublist);

  error=zxs_addChunkToList(sub,"SNAP"); /* Form type of the RIFF chunk */
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  error=zxs_addChunkToList(sub,"fmtz");
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  error=zxs_addChunkToList(sub,"rZ80");
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  error=zxs_addChunkToList(sub,"r048");
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  error=zxs_addChunkToList(sub,"ram0");
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  error=zxs_addChunkToList(sub,"ram2");
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  zxs_addChunkToList(sub,"ram5");
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  if(sptr->machine!=MACHINE_48K) /* 128K machines */
    {
      zxs_addChunkToList(sub,"r128");
      if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

      zxs_addChunkToList(sub,"ram1");
      if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

      zxs_addChunkToList(sub,"ram3");
      if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

      zxs_addChunkToList(sub,"ram4");
      if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

      zxs_addChunkToList(sub,"ram6");
      if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

      zxs_addChunkToList(sub,"ram7");
      if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

      if(sptr->machine==MACHINE_PLUS2A || sptr->machine==MACHINE_PLUS3)
	{
	  zxs_addChunkToList(sub,"r+3 ");
	  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }
	}
    }

  error=zxs_writeChunk(f,sptr,pages,head,&length);
  if(error!=ERROR_NONE) { zxs_deleteList(&head); return error; }

  zxs_deleteList(&head);

  return ERROR_NONE;

}

/* Overall +D control functions */

int plusd_read2(FILE *f,char *inName,struct snapinfo *sptr,
		unsigned char **pages,romData *romPages)
{int error; unsigned char *buffer;

 if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
   return ERROR_MEMORY;

 error=plusd_read(f,sptr,buffer);
 if(error!=ERROR_NONE) { free(buffer); return error; }
 if(verbose>=1) fprintf(stderr,"+D snap read successfully\n");

 error=construct48KPages(buffer,pages);
 if(error!=ERROR_NONE) { free(buffer); return error; }

 free(buffer);

 return ERROR_NONE;

}

int plusd_write2(FILE *f,char *outName,int version,struct snapinfo *sptr,
		 unsigned char **pages,romData *romPages)
{
  unsigned char *buffer; int error;

  error=unsupportedHardware(sptr,UNSUPPORTED_48KNOPERIPHS,"+D");
  if(error!=ERROR_NONE) return error;

  if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==NULL)
    return ERROR_MEMORY;

  construct48KBuffer(pages,buffer);

  error=plusd_write(f,sptr,buffer);
  if(error!=ERROR_NONE) { free(buffer); return error; }
  if(verbose>=1) fprintf(stderr,"+D snap written successfully\n");

  free(buffer);

  return ERROR_NONE;

}

/* Utility routines */

int printError(char *string,int errorNum)
{
  if(string!=0)
    {
      fprintf(stderr,"%s: ",string);
    }

  if(errorNum<numErrorMessages)
    {
      fprintf(stderr,"%s\n",errorMessages[errorNum]);
      return 1;
    }
  else
    {
      fprintf(stderr,"Internal error: unknown error number %d\n",errorNum);
      return 255;
    }
}

int construct48KBuffer(unsigned char **pages,unsigned char *buffer)
{
 if(pages[5]) memcpy(&buffer[0x0000],pages[5],0x4000);
   else memset(&buffer[0x0000],0,0x4000);
 if(pages[2]) memcpy(&buffer[0x4000],pages[2],0x4000);
   else memset(&buffer[0x4000],0,0x4000);
 if(pages[0]) memcpy(&buffer[0x8000],pages[0],0x4000);
   else memset(&buffer[0x8000],0,0x4000);

 return ERROR_NONE;

}

int construct48KPages(unsigned char *buffer, unsigned char **pages)
{
 pages[5]=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
 if(!pages[5]) {                                return ERROR_MEMORY;}
 pages[2]=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
 if(!pages[2]) {                free(pages[5]); return ERROR_MEMORY;}
 pages[0]=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
 if(!pages[0]) {free(pages[2]); free(pages[5]); return ERROR_MEMORY;}

 memcpy(pages[5],&buffer[0x0000],0x4000);
 memcpy(pages[2],&buffer[0x4000],0x4000);
 memcpy(pages[0],&buffer[0x8000],0x4000);

 return ERROR_NONE;

}

void writeInt(int i,FILE *f)
{fputc(i%0x100,f); fputc(i/0x100,f); return;}

#ifndef HAVE_STRCASECMP

/* Some crippled systems don't define this... */

int strcasecmp (const char *S1, const char *S2)
{
  char c1,c2;

  for( ; *S1 && *S2 ; S1++, S2++ )
    {
      c1=tolower(*S1); c2=tolower(*S2);
      if( c1 != c2 ) return ( c1 - c2 );
    }

  return ( tolower(*S1) - tolower(*S2) );

}

#endif /* #ifndef HAVE_STRCASECMP */
