/* sp.c: routines for dealing with .sp format files
   Copyright (c) 1997-8 Philip Kendall

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: pak21@cam.ac.uk (until at least June 1999)
   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England

*/

#include <stdio.h>
#include <string.h>

#include "snapconv.h"
#include "snapinfo.h"

#include "sp.h"

int sp_read(FILE *f,struct snapinfo *sptr,unsigned char *buffer)
{
  unsigned char Header[sp_headerSize];
  unsigned length,start;

  /* Read in the first six bytes to see if this is an old
    or a new .SP file */
  if(fread(Header,6,1,f)!=1) return ERROR_READ;

  if(Header[0]=='S' && Header[1]=='P') /* New style */
    {
      if(verbose>=2) fprintf(stderr,"Reading a new style .sp file\n");
      if(fread(&Header[6],sp_headerSize-6,1,f)!=1) return ERROR_READ;
    }
  else
    {
      memcpy(&Header[6],&Header[0],6); /* Move the bytes up */
      memcpy(&Header[0],"SP\x00\xc0\x00\x40",6); /* Add new header bytes */

      if(verbose>=2) fprintf(stderr,"Reading an old style .sp file\n");

      /* and read in the rest of the header */
      if(fread(&Header[12],sp_headerSize-12,1,f)!=1) return ERROR_READ;
    }

  /* These two interpreted only for use later */
  length=Header[2]+Header[3]*0x100;
  start=Header[4]+Header[5]*0x100;

  sptr->bc=Header[ 6]+Header[ 7]*0x100;
  sptr->de=Header[ 8]+Header[ 9]*0x100;
  sptr->hl=Header[10]+Header[11]*0x100;
  sptr->f=Header[12]; sptr->a=Header[13];

  sptr->ix=Header[14]+Header[15]*0x100;
  sptr->iy=Header[16]+Header[17]*0x100;

  sptr->bc_=Header[18]+Header[19]*0x100;
  sptr->de_=Header[20]+Header[21]*0x100;
  sptr->hl_=Header[22]+Header[23]*0x100;
  sptr->f_=Header[24]; sptr->a_=Header[25];

  sptr->r=Header[26]; sptr->i=Header[27];

  sptr->sp=Header[28]+Header[29]*0x100;
  sptr->pc=Header[30]+Header[31]*0x100;

  sptr->outULA=Header[34]&0x07;

  sptr->iff1=Header[36]&0x01;
  sptr->im= Header[36]&0x08 ? 0 : ( (Header[36]&0x02) >> 1 ) + 1;
  sptr->iff2=(Header[36]&0x04)>>2;
  sptr->interrupt=(Header[36]&0x10)>>4;
  sptr->flash=(Header[36]&0x20)>>5;

  if(verbose>=2)
    {
      fprintf(stderr,"BC =0x%04x DE =0x%04x\n",sptr->bc,sptr->de);
      fprintf(stderr,"HL =0x%04x AF =0x%02x%02x\n",sptr->hl,sptr->a,sptr->f);
      fprintf(stderr,"IX =0x%04x IY =0x%04x\n",sptr->ix,sptr->iy);
      fprintf(stderr,"BC'=0x%04x DE'=0x%04x\n",sptr->bc_,sptr->de_);
      fprintf(stderr,"HL'=0x%04x AF'=0x%02x%02x\n",
	      sptr->hl_,sptr->a_,sptr->f_);
      fprintf(stderr," R =0x%02x    I =0x%02x\n",sptr->r,sptr->i);
      fprintf(stderr,"SP =0x%04x PC =0x%04x\n",sptr->sp,sptr->pc);
      fprintf(stderr,"Border colour %d\n",sptr->outULA);
      fprintf(stderr,"IFF1=%d IFF2=%d IM %d\n",sptr->iff1,sptr->iff2,sptr->im);
      fprintf(stderr,"%snterrupt pending\n",(sptr->interrupt ? "I" : "No i"));
      fprintf(stderr,"Flash is %s\n",(sptr->flash ? "paper" : "ink"));
    }

  memset(buffer,0,0xc000);	/* zero memory contents */

  if(fread(&buffer[start-0x4000],length,1,f)!=1) return ERROR_READ;

  if(verbose>=2) fprintf(stderr,"Read in 0x%04x bytes starting at 0x%04x\n",
			 length,start);

  return ERROR_NONE;

}

int sp_write(FILE *f,struct snapinfo *sptr,int version,unsigned char *buffer)
{int temp1,temp2;

 if(version>=2)
   {
     fputs("SP",f);		/* signature */
     writeInt(0xc000,f);	/* program length */
     writeInt(0x4000,f);	/* program location */
   }

 writeInt(sptr->bc,f); writeInt(sptr->de,f);
 writeInt(sptr->hl,f); fputc(sptr->f,f); fputc(sptr->a,f);

 writeInt(sptr->ix,f); writeInt(sptr->iy,f);

 writeInt(sptr->bc_,f); writeInt(sptr->de_,f);
 writeInt(sptr->hl_,f); fputc(sptr->f_,f); fputc(sptr->a_,f);

 fputc(sptr->r,f); fputc(sptr->i,f);

 writeInt(sptr->sp,f); writeInt(sptr->pc,f);

 writeInt(0,f);		/* reserved */

 fputc( (sptr->outULA)&0x07 , f);

 fputc(0,f);		/* reserved */

 temp1 = (sptr->im)==0;
 temp2 = temp1 ? 0 : (sptr->im)-1 ;

 writeInt(sptr->iff1 + ( temp2 << 1) + (sptr->iff2 << 2) + ( temp1 << 3 ) +
	  (sptr->interrupt << 4) + (sptr->flash << 5) ,f);

 /* and finally write out memory contents */

 fwrite(buffer,0xc000,1,f);

 return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}
