/* z80.c: routines for dealing with .z80 format files
   Copyright (c) 1997-8 Philip Kendall

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: pak21@cam.ac.uk (until at least June 1999)
   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England

*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#include "hardware.h"
#include "snapconv.h"
#include "snapinfo.h"

#include "z80.h"		/* prototypes for the public routines */

static void doRepeat(unsigned char **ptr,const unsigned char **src,
		     const unsigned char *End);
static void compressED(unsigned char **ptr,const unsigned char **src,
		       const unsigned char *End);

unsigned z80_compress(unsigned char *dest,const unsigned char *src,
		      unsigned Length)
{const unsigned char *End; unsigned char *ptr;
 End=src+Length;		/* End points to 1 past end of src */
 ptr=dest;
 while(src<(End-1))
     if(*src==*(src+1)) doRepeat(&ptr,&src,End);
     else { *ptr++=*src++; if(*(src-1)==(unsigned char)'\xed') *ptr++=*src++; }
 /* If the last character hasn't been taken into a block, copy it across */
 if(src==End-1) *ptr++=*src++;
 return (ptr-dest);
}

static void doRepeat(unsigned char **ptr,const unsigned char **src,
		     const unsigned char *End)
{unsigned char ch; unsigned RunLength;
 /* filter out EDs */
 if(**src==(unsigned char)'\xed') compressED(ptr,src,End);
 /* deal with everything else - first, if we're near the end, just copy */
 else if(*src>=End-4) while(*src<End) { **ptr=**src; (*ptr)++; (*src)++; }
 /* see if we've got a run of at least 5; we already know that we've got
    a run of two, so no need to check *((*ptr)+1) */
 else
 {ch=**src; (*src)+=2; RunLength=2;
  while(**src==ch && RunLength<0xff && (*src)<End)
      {RunLength++; (*src)++;}
  if(RunLength>=5)
  {/* Write out 2 EDs */
   **ptr=(unsigned char)'\xed'; (*ptr)++; **ptr=(unsigned char)'\xed'; (*ptr)++;
   /* Write out length and byte */
   **ptr=(unsigned char)RunLength; (*ptr)++;
   **ptr=ch; (*ptr)++;
  }
  else				/* write out the bytes directly */
  while(RunLength--) {**ptr=ch; (*ptr)++;}
 }
 return;
}

static void compressED(unsigned char **ptr,const unsigned char **src,
		       const unsigned char *End)
{unsigned char ch; unsigned RunLength;
 ch=(unsigned char)'\xed'; (*src)+=2; RunLength=2;
 while(**src==ch && RunLength<0xff && (*src)<End)
     {RunLength++; (*src)++;}
 **ptr=ch; (*ptr)++; **ptr=ch; (*ptr)++;
 **ptr=(unsigned char)RunLength; (*ptr)++; **ptr=ch; (*ptr)++;
 return;
}

unsigned z80_uncompress(unsigned char *dest,const unsigned char *src,
			unsigned Length)
{unsigned char *ptr=dest,ch; const unsigned char *End=src+Length;
 unsigned RunLength;
 while(src<End)
 {if(*src==(unsigned char)'\xed' && *(src+1)==(unsigned char)'\xed')
  {src+=2; RunLength=*src++; ch=*src++; while(RunLength--) *ptr++=ch;}
  else *ptr++=*src++;
 }
 return (ptr-dest);
}

int z80_readHeader(FILE *f,struct snapinfo *sptr)
{unsigned char Header[z80_v1_headerLength],*NewHeader=NULL;
 unsigned long temp; unsigned NewLength;

 if( fread(Header,z80_v1_headerLength,1,f)!=1 ) return ERROR_READ;

 sptr->a=Header[0]; sptr->f=Header[1];
 sptr->bc=Header[ 2]+Header[ 3]*0x100;
 sptr->de=Header[13]+Header[14]*0x100;
 sptr->hl=Header[ 4]+Header[ 5]*0x100;
 sptr->a_=Header[21]; sptr->f_=Header[22];
 sptr->bc_=Header[15]+Header[16]*0x100;
 sptr->de_=Header[17]+Header[18]*0x100;
 sptr->hl_=Header[19]+Header[20]*0x100;
 sptr->ix=Header[25]+Header[26]*0x100;
 sptr->iy=Header[23]+Header[24]*0x100;
 sptr->i=Header[10]; sptr->r=(Header[11]&0x7f)+((Header[12]&0x01)<<7);
 sptr->pc=Header[6]+Header[7]*0x100;
 sptr->sp=Header[8]+Header[9]*0x100;

 /* `!!' is to reduce possible values to 0 or 1 */
 sptr->iff1=!!Header[27]; sptr->iff2=!!Header[28];
 sptr->im=Header[29]&0x03;

 sptr->outULA=(Header[12]&0x0e)>>1;
 sptr->issue2=(Header[29]&0x04)>>2;
 sptr->doubleint=(Header[29]&0x08)>>3;
 sptr->videosync=(Header[29]&0x2d)>>4;
 sptr->joystick=(Header[29]&0xc0)>>6;

 if(verbose>=2)
   {
     const char *ptr;

     fprintf(stderr,"AF =0x%02x%02x BC =0x%04x\n",sptr->a,sptr->f,
	     sptr->bc);
     fprintf(stderr,"DE =0x%04x HL =0x%04x\n",sptr->de,sptr->hl);
     fprintf(stderr,"AF'=0x%02x%02x BC'=0x%04x\n",sptr->a_,sptr->f_,
	     sptr->bc_);
     fprintf(stderr,"DE'=0x%04x HL'=0x%04x\n",sptr->de_,sptr->hl_);
     fprintf(stderr,"IX =0x%04x IY =0x%04x\n",sptr->ix,sptr->iy);
     fprintf(stderr," I =0x%02x    R =0x%02x\n",sptr->i,sptr->r);
     fprintf(stderr,"PC =0x%04x (0x0000 => v2+ .z80 file)\n",
	     sptr->pc);
     fprintf(stderr,"SP =0x%04x\n",sptr->sp);
     fprintf(stderr,"IFF1=%d IFF2=%d IM %d\n",sptr->iff1,sptr->iff2,
	     sptr->im);
     fprintf(stderr,"Border colour: %d\n",sptr->outULA);
     fprintf(stderr,"Issue 2 emulation: %s\n",
	     (sptr->issue2 ? "On" : "Off"));
     fprintf(stderr,"Double interrupt frequency: %s\n",
	     (sptr->doubleint ? "On" : "Off"));
     fprintf(stderr,"Video syncronisation: %s\n",
	     ( (sptr->videosync==0) || (sptr->videosync==2)
	       ? "Normal"
	       : (sptr->videosync==1) ? "High" : "Low"
	       )
	     );
     switch(sptr->joystick)
       {
	 case 0 : ptr="Cursor"; break;
	 case 1 : ptr="Kempston"; break;
	 case 2 : ptr="Left Sinclair/User defined"; break;
	 case 3 : ptr="Right Sinclair"; break;
	 default: ptr="Internal error: unknown!"; break;
       }
     fprintf(stderr,"Joystick emulation: %s\n",ptr);
   }

 if(!sptr->pc)			/* PC==0 => v2 or greater .z80 file */
   {
     NewLength=fgetc(f)+fgetc(f)*0x100;
     if(ferror(f)) return ERROR_READ;

     if(verbose>=2) fprintf(stderr,"Extra header length: %d\n",NewLength);

     switch(NewLength)
       {
	 case z80_v2_headerLength: sptr->version=2; break;
	 case z80_v3_headerLength:
	 case z80_v3x_headerLength: sptr->version=3; break;
	 default: return ERROR_UNKNOWNHEADERLEN;
       }

     if(verbose>=2) fprintf(stderr,"=> .z80 version %d\n",sptr->version);

     NewHeader=(unsigned char*)malloc(NewLength*sizeof(unsigned char));
     if(!NewHeader) return ERROR_MEMORY;

     fread(NewHeader,NewLength,1,f);
     if(ferror(f)) return ERROR_READ;

     sptr->pc=NewHeader[0]+NewHeader[1]*0x100;
     sptr->rreg=NewHeader[5]&0x01;
     sptr->ldir=(NewHeader[5]&0x02)>>1;

     if(sptr->version<3)
       {
	 switch(NewHeader[2])
	   {
	     case 0 : sptr->machine=MACHINE_48K;  sptr->periph=PERIPH_NONE;
	              break;
	     case 1 : sptr->machine=MACHINE_48K;  sptr->periph=PERIPH_IF1;
	              break;
	     case 2 : sptr->machine=MACHINE_48K;  sptr->periph=PERIPH_SAMRAM;
	              break;
	     case 3 : sptr->machine=MACHINE_128K; sptr->periph=PERIPH_NONE;
	              break;
	     case 4 : sptr->machine=MACHINE_128K; sptr->periph=PERIPH_IF1;
	              break;
	     default: return ERROR_UNKNOWNHARDWARE;
	   }
       }
     else
       {
	 switch(NewHeader[2])
	   {
	     case 0 : sptr->machine=MACHINE_48K;   sptr->periph=PERIPH_NONE;
	              break;
	     case 1 : sptr->machine=MACHINE_48K;   sptr->periph=PERIPH_IF1;
	              break;
	     case 2 : sptr->machine=MACHINE_48K;   sptr->periph=PERIPH_SAMRAM;
	              break;
	     case 3 : sptr->machine=MACHINE_48K;   sptr->periph=PERIPH_MGT;
	              break;
	     case 4 : sptr->machine=MACHINE_128K;  sptr->periph=PERIPH_NONE;
	              break;
	     case 5 : sptr->machine=MACHINE_128K;  sptr->periph=PERIPH_IF1;
	              break;
	     case 6 : sptr->machine=MACHINE_128K;  sptr->periph=PERIPH_MGT;
	              break;
	     case 7 : sptr->machine=MACHINE_PLUS3; sptr->periph=PERIPH_NONE;
	              break;
	     case 8 : sptr->machine=MACHINE_PLUS3; sptr->periph=PERIPH_IF1;
	              break;
	     default: return ERROR_UNKNOWNHARDWARE;
	   }
       }

     if(verbose>=2)
       {
	 const char *ptr;

	 fprintf(stderr,"PC is actually 0x%04x\n",sptr->pc);
	 fprintf(stderr,"R register emulation: %s\n",
		 (sptr->rreg ? "On" : "Off"));
	 fprintf(stderr,"LDIR emulation: %s\n",
		 (sptr->ldir ? "On" : "Off"));


	 switch(sptr->machine)
	   {
	     case MACHINE_48K  : ptr="48k";  break;
	     case MACHINE_128K : ptr="128k"; break;
	     case MACHINE_PLUS3: ptr="+3";   break;
	     default           : ptr="Internal error: unknown!"; break;
	   }
	 fprintf(stderr,"Basic hardware: %s\nPeripherals: ",ptr);
	 if(sptr->periph==PERIPH_NONE) fprintf(stderr,"None\n");
	 else
	   {
	     if(sptr->periph&PERIPH_IF1   ) fprintf(stderr,"Interface 1 ");
	     if(sptr->periph&PERIPH_MGT   ) fprintf(stderr,"MGT ");
	     if(sptr->periph&PERIPH_SAMRAM) fprintf(stderr,"SamRam ");
	     fprintf(stderr,"\n");
	   }
       }


     if( (sptr->periph)&PERIPH_SAMRAM)	/* SamRam present */
       {
	 sptr->samram=NewHeader[3];
	 if(verbose>=2) fprintf(stderr,"SamRam latch state: 0x%02x\n",
				sptr->samram);
       }

     if( (sptr->machine)!=MACHINE_48K )	/* 128k modes */
       {
	 /* If we have an xzx +3 snap, get the last out to 0x1ffd */
	 if(NewLength==z80_v3x_headerLength) sptr->out1ffd=NewHeader[54];

	 sptr->out7ffd=NewHeader[3];

	 sptr->ayactive=1;
	 sptr->outfffd=NewHeader[6];
	 memcpy(sptr->soundchip,&NewHeader[7],16);

	 if(verbose>=2)
	   {
	     if(NewLength==z80_v3x_headerLength)
	       fprintf(stderr,"Last OUT to 0x1ffd: 0x%02x\n",sptr->out1ffd);
	     fprintf(stderr,"Last OUT to 0x7ffd: 0x%02x\n",sptr->out7ffd);
	   }
       }
     else
       {
	 sptr->ayactive=(NewHeader[5]&0x04)>>2;

	 if(verbose>=2)
	   {
	     fprintf(stderr,"AY-3-8912 is %s\n",
		     (sptr->ayactive ? "active" : "inactive"));
	   }
       }

     if(sptr->ayactive)
       {
	 sptr->outfffd=NewHeader[6];
	 memcpy(sptr->soundchip,&NewHeader[7],16);

	 if(verbose>=2)
	   {

	     int i;

	     fprintf(stderr,"Last OUT to 0xfffd: 0x%02x\n",sptr->outfffd);
	     fprintf(stderr,"Soundchip registers:");
	     for(i=0;i<16;i++)
	       fprintf(stderr," 0x%02x%s",(unsigned)sptr->soundchip[i],
		       (i==15 ? "\n" : (i==7 ? ",\n                    "
					: ","
				       )
		       )
		      );
	   }
       }

     sptr->if1= (NewHeader[4]==0xff) ;
     if(verbose>=2) fprintf(stderr,"Interface 1 ROM is %spaged\n",
			    (sptr->if1 ? "" : "not " ));

     if(sptr->version>=3)		/* read the extra v3 headers */
       {
	 sptr->mgt= (NewHeader[27]==0xff) ;
	 sptr->multiface= (NewHeader[28]==0xff) ;
	 if(verbose>=2)
	   {
	     fprintf(stderr,"MGT ROM is %spaged\n",
		     (sptr->mgt ? "" : "not " ));
	     fprintf(stderr,"Multiface ROM is %spaged\n",
		     (sptr->multiface ? "" : "not " ));
	   }

	 if( (sptr->periph)&PERIPH_MGT ) /* MGT modes */
	   {
	     sptr->mgttype=NewHeader[51];
	     sptr->discbutton= (NewHeader[52]==0xff) ;
	     sptr->discflag= (NewHeader[53]==0xff) ;

	     if(verbose>=2)
	       {
		 const char *ptr;
		 switch(sptr->mgttype)
		   {
		      case 0: ptr="Disciple + Epson"; break;
		      case 1: ptr="Disciple + HP"; break;
		     case 16: ptr="+D"; break;
		     default: ptr="Unknown!"; break;
		   }
		 fprintf(stderr,"MGT type: %s\n",ptr);
		 fprintf(stderr,"Disciple inhibit button: %s\n",
			 (sptr->discbutton ? "In" : "Out" ));
		 fprintf(stderr,"Disciple inhibit flag: ROM %spageable\n",
			 (sptr->discflag ? "not " : "" ));
	       }

	   }

	 temp=(sptr->machine==MACHINE_48K)?17472:17727;
	 /* This number is 1/4 of an interrupt cycle */

	 /* Thanks to Pedro Gimeno for pointing out I was one off here
	    (even if it did agree with z80 v3.05's Z80DUMP...) */
	 sptr->tstates=(((NewHeader[25]+1)%4)+1)*temp
	   -1-NewHeader[23]-NewHeader[24]*0x100;

	 sptr->spectator=NewHeader[26];

	 memcpy(sptr->userstick,&NewHeader[31],20);

	 if(verbose>=2)
	   {
	     int i;

	     fprintf(stderr,"T-States since interrupt: %ld\n",
		     sptr->tstates);

	     fprintf(stderr,"Spectator flag byte: 0x%02x\n",
		     sptr->spectator);

	     fprintf(stderr,"User-defined joystick keyboard mappings:\n");
	     for(i=0;i<5;i++)
	       fprintf(stderr,"  Row %d, column 0x%02x (`%c')\n",
		       (unsigned)sptr->userstick[i*2],
		       (unsigned)sptr->userstick[i*2+1],
		       sptr->userstick[i*2+10]);
	   }
       }

     free(NewHeader);

   }
 else			/* v1 .z80 file */
   {
     sptr->version=1;
     sptr->compressed=(Header[12]&0x20)>>5;

     /* If SamRom is not switched in, assume that it's not present */

     if(Header[12]&0x10)
       {
	 sptr->periph&=PERIPH_SAMRAM;
	 sptr->samram=0x5995;
       }

     if(verbose>=2)
       {
	 fprintf(stderr,"Data is %scompressed\n",
		 (sptr->compressed ? "" : "not "));
	 fprintf(stderr,"SamRam is %spresent",
		 (sptr->periph&PERIPH_SAMRAM ? "" : "not "));
       }

   }

 return ERROR_NONE;

}

int z80_writev1Header(FILE *f,const struct snapinfo *sptr,int v2)
{

  fputc(sptr->a,f); fputc(sptr->f,f);
  writeInt(sptr->bc,f); writeInt(sptr->hl,f);

  writeInt( (v2 ? 0 : sptr->pc) ,f); /* PC==0 => v2 or greater .z80 file */

  writeInt(sptr->sp,f); fputc(sptr->i,f);
  fputc((sptr->r)&0x7f,f);

  /* assume the data is compressed; we may have to reset bit 5 later on
     if it's not	*/
  fputc((int)( ((sptr->r)>>7) + (( (sptr->outULA)&0x07 )<<1)
	 + ( v2 ? 0 : (((sptr->samram)&0x80)>>3) + 0x20 ) ),f);

  writeInt(sptr->de ,f); writeInt(sptr->bc_,f);
  writeInt(sptr->de_,f); writeInt(sptr->hl_,f);
  fputc(sptr->a_,f); fputc(sptr->f_,f);
  writeInt(sptr->iy,f); writeInt(sptr->ix,f);

  fputc( (sptr->iff1 ? 0xff : 0x00 ) ,f);
  fputc( (sptr->iff2 ? 0xff : 0x00 ) ,f);

  fputc( sptr->im + ((sptr->issue2)<<2) + ((sptr->doubleint)<<3) +
	 ((sptr->videosync)<<4) + ((sptr->joystick)<<6) ,f);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int z80_writev2Header(FILE *f,const struct snapinfo *sptr,int v3)
{
 int error; int length;

 /* write out the v1 header */

 if((error=z80_writev1Header(f,sptr,1))!=0) return error;

 /* and follow that with the v2 header */

 if(v3)
   {
     length = (sptr->machine==MACHINE_PLUS3 ? z80_v3x_headerLength :
	       z80_v3_headerLength);
   }
 else length=z80_v2_headerLength;

 writeInt(length,f);
 writeInt(sptr->pc,f);

 /* Deal with the hardware byte differently for v2 and v3 */

 if(v3)
   {
     if(sptr->machine==MACHINE_PLUS3)
       {
	      if( (sptr->periph)&PERIPH_IF1 ) fputc(8,f);
	 else				      fputc(7,f);
       }
     else if(sptr->machine==MACHINE_128K)
       {
	      if( (sptr->periph)&PERIPH_MGT ) fputc(6,f);
	 else if( (sptr->periph)&PERIPH_IF1 ) fputc(5,f);
	 else				      fputc(4,f);
       }
     else if(sptr->machine==MACHINE_48K)
       {
	      if( (sptr->periph)&PERIPH_SAMRAM ) fputc(2,f);
	 else if( (sptr->periph)&PERIPH_MGT )    fputc(3,f);
	 else if( (sptr->periph)&PERIPH_IF1 )    fputc(1,f);
	 else				         fputc(0,f);
       }
   }
 else
   {
     if(sptr->machine!=MACHINE_48K)
       {
	 if( (sptr->periph)&PERIPH_IF1 ) fputc(4,f);
	 else  			         fputc(3,f);
       }
     else
       {
	      if( (sptr->periph)&PERIPH_SAMRAM ) fputc(2,f);
	 else if( (sptr->periph)&PERIPH_IF1 )    fputc(1,f);
	 else				         fputc(0,f);
       }
   }

 if( (sptr->machine)!=MACHINE_48K ) /* 128K modes */
   {
     fputc(sptr->out7ffd,f);
   }
 else if( (sptr->periph)&PERIPH_SAMRAM ) /* SamRam mode */
   {
     fputc((int)(sptr->samram),f);
   }
 else
   {
     fputc(0,f);
   }

 fputc( (sptr->if1 ? 0xff : 0x00 ) ,f);
 fputc( sptr->rreg + ((sptr->ldir)<<1) + ((sptr->ayactive)<<2) ,f);
 fputc(sptr->outfffd,f);
 fwrite(sptr->soundchip,16,1,f);

 return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int z80_writev3Header(FILE *f,const struct snapinfo *sptr)
{int error,temp;

 /* write out all the v1 and v2 headers */

 if((error=z80_writev2Header(f,sptr,1))!=0) return error;

 /* and then the extra v3 headers */

 temp=( (sptr->machine)>=1) ? 17727 : 17472;
 writeInt((int)( temp - ((sptr->tstates)%temp) - 1 ),f);
 fputc( (int)((((sptr->tstates)/temp)+3)%4),f);

 fputc(sptr->spectator,f);
 fputc( ( sptr->mgt       ? 0xff : 0x00 ) ,f);
 fputc( ( sptr->multiface ? 0xff : 0x00 ) ,f);

 /* 0x0000 to 0x1fff is RAM is we're in a +3 special configuration */

 if( sptr->machine==MACHINE_PLUS3 && (sptr->out1ffd & 0x01) )
   temp=0x00;
 else
   temp=0xff;

 fputc(temp,f);		/* 0x0000 to 0x1fff is always ROM */

 /* 0x2000 to 0x3fff is RAM if Multiface is paged, or if we're in
    a +3 special configuration */

 fputc( ( !temp || sptr->multiface ? 0x00 : 0xff ) , f );

 fwrite(sptr->userstick,20,1,f);
 fputc(sptr->mgttype,f);
 fputc( ( sptr->discbutton ? 0xff : 0x00 ) ,f);
 fputc( ( sptr->discflag   ? 0xff : 0x00 ) ,f);

 if(sptr->machine==MACHINE_PLUS3) fputc( sptr->out1ffd , f);

 return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int z80_readBlock(FILE *f,int *pptr,unsigned char **buffer,
		  unsigned *uncompLength)
{

  unsigned compLength; unsigned char *buffer2;

  compLength=fgetc(f)+fgetc(f)*0x100; *pptr=fgetc(f);
  if(feof(f)) return 0;
  if(ferror(f)) return ERROR_READ;

  (*buffer)=(unsigned char*)malloc(0x4000*sizeof(unsigned char));
  if(*buffer==NULL) return ERROR_MEMORY;

  if(compLength!=0xffff) /* length==0xffff => uncompressed block */
    {
      buffer2=(unsigned char*)malloc(0x7000*sizeof(unsigned char));
      if(!buffer2) { free(buffer); return ERROR_MEMORY; }

      fread(buffer2,compLength,1,f);
      if(ferror(f)) { free(buffer2); free(buffer); return ERROR_READ; }

      (*uncompLength)=z80_uncompress(*buffer,buffer2,compLength);

      /* Pages should not expand to more than 16K */

      if(*uncompLength>0x4000)
	{
	  free(buffer2); free(buffer);
	  return ERROR_CORRUPT;
	}

      /* The IF1 and Multiface ROMs are only 8K long */

      if(*uncompLength<0x4000)
	{
	  *buffer=realloc(buffer,*uncompLength);
	  if(*buffer==NULL) { free(buffer2); return ERROR_MEMORY; }
	}

      free(buffer2);

    }
  else
    {
      fread(*buffer,0x4000,1,f);
      if(ferror(f)) { free(buffer); return ERROR_READ; }
      (*uncompLength)=0x4000;
    }

  return ERROR_NONE;

}

int z80_readv1Data(FILE *f,unsigned char *buffer,int compressed)
{unsigned char *buffer2,*ptr;

 if(compressed)
 {
  buffer2=(unsigned char*)malloc(0xd000*sizeof(unsigned char));
  if(!buffer2) return ERROR_MEMORY;

  ptr=buffer2;
  while(!feof(f) && !ferror(f)) *ptr++=fgetc(f);
  if(ferror(f)) {free(buffer2); return ERROR_READ;}

  /* -3 is +1 to count both ends and -4 for the end marker */
  z80_uncompress(buffer,buffer2,(unsigned int)((ptr-buffer2)-3));

  free(buffer2);

 }
 else
 {fread(buffer,0xc000,1,f);
  if(ferror(f)) return ERROR_READ;
 }

 return 0;

}

int z80_writeBlock(FILE *f,int num,unsigned char *page,unsigned long length)
{unsigned char *buffer; int compressedLength;

 if((buffer=(unsigned char*)malloc(0x7000*sizeof(unsigned char)))==NULL)
	return ERROR_MEMORY;

 compressedLength=z80_compress(buffer,page,length);

 if((unsigned long)compressedLength<=length)
 {
   writeInt(compressedLength,f); fputc(num,f);
   fwrite(buffer,(size_t)compressedLength,1,f);
 }
 else
 {
   writeInt(0xffff,f); fputc(num,f);
   fwrite(page,(size_t)length,1,f);
 }

 free(buffer);

 return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int z80_writev1Data(FILE *f,unsigned char *page5,unsigned char *page1,
		    unsigned char *page2)
{unsigned char *buffer,*buffer2; unsigned length;

 if((buffer=(unsigned char*)malloc(0xc000*sizeof(unsigned char)))==0)
	return ERROR_MEMORY;
 if((buffer2=(unsigned char*)malloc(0xf000*sizeof(unsigned char)))==0)
	{free(buffer); return ERROR_MEMORY;}

 if(page5) memcpy(&buffer[0x0000],page5,0x4000);
   else memset(&buffer[0x0000],0,0x4000);
 if(page1) memcpy(&buffer[0x4000],page1,0x4000);
   else memset(&buffer[0x4000],0,0x4000);
 if(page2) memcpy(&buffer[0x8000],page2,0x4000);
   else memset(&buffer[0x8000],0,0x4000);

 length=z80_compress(buffer2,buffer,0xc000);

 if(length<0xc000)
 {fwrite(buffer2,length,1,f);
  fwrite("\0\xed\xed\0",4,1,f);	/* end marker */
 }
 else
 {fwrite(buffer,0xc000,1,f);
  fseek(f,12,SEEK_SET);
  fputc(fgetc(f)^0x20,f);	/* reset the 'compressed' bit */
  fseek(f,0,SEEK_END);
 }

 free(buffer); free(buffer2);

 return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int z80_parsePage(int page,unsigned char *buffer,unsigned length,
		  struct snapinfo *sptr,unsigned char **pages,
		  romData *romPages)
{

  /* Deal with any ROMs first; we have to be careful with page number
     3 as it can also be 128K RAM page 0
  */

  switch(page)
    {
      case 0:

	if(sptr->machine==MACHINE_128K)
	  return z80_parseRom(buffer,length,ROM_128K_1,romPages);

	else
	  return z80_parseRom(buffer,length,ROM_48K,romPages);

      case 1:

	if(sptr->periph&PERIPH_IF1)
	  return z80_parseRom(buffer,length,ROM_IF1,romPages);

	else if(sptr->periph&PERIPH_MGT)
	  {
	    switch(sptr->mgttype)
	      {
		case 0:
		case 1:
		  return z80_parseRom(buffer,length,ROM_DISCIPLE,romPages);
		case 16:
		  return z80_parseRom(buffer,length,ROM_PLUSD,romPages);
	      }
	  }

	else { free(buffer); return ERROR_UNKNOWNPAGE; }

      case 2:

	if(sptr->machine==MACHINE_128K)
	  return z80_parseRom(buffer,length,ROM_128K_0,romPages);

	else if(sptr->periph&PERIPH_SAMRAM)
	  return z80_parseRom(buffer,length,ROM_SAMRAM_0,romPages);

	else { free(buffer); return ERROR_UNKNOWNPAGE; }

      case 3:

	if(sptr->periph&PERIPH_SAMRAM)
	  return z80_parseRom(buffer,length,ROM_SAMRAM_1,romPages);

	else if(sptr->machine==MACHINE_48K)
	  { free(buffer); return ERROR_UNKNOWNPAGE; }

	/* We must pass page 3 on as it can also be 128K RAM bank 0 */

	break;

      case 11:

	if( ! (sptr->periph&PERIPH_SAMRAM) )
	  return z80_parseRom(buffer,length,ROM_MULTIFACE,romPages);

	else { free(buffer); return ERROR_UNKNOWNPAGE; }
    }

  if(page<3 || page>10) return ERROR_UNKNOWNPAGE;

  page-=3;

   /* Renumber the 48K pages according to how I like them :-) */

  if(sptr->machine==MACHINE_48K)
    {
      switch(page)
	{
	  /* Change from Gerton's numbering scheme to that used by
	     the 128K Speccy */

	  case 0: page=1; break;
	  case 1: page=2; break;
	  case 2: page=0; break;

	}
    }

  if(pages[page]) return ERROR_DUPLICATEPAGE;

  pages[page]=buffer;

  return ERROR_NONE;

}

int z80_parseRom(unsigned char *buffer,unsigned length,int which,
		 romData *romPages)
{
  if(romPages[which].present)
    {
      free(buffer); return ERROR_DUPLICATEPAGE;
    }

  if(length!=romInfo[which].length)
    {
      free(buffer); return ERROR_PAGELENGTH;
    }

  romPages[which].present=1;

#ifdef HAVE_LIBZ
  if( adler32( adler32(0L,Z_NULL,0) , buffer, length) !=
      romInfo[which].checksum)
    {
      romPages[which].ptr=buffer; return ERROR_NONE;
    }
  else
#endif /* #ifdef HAVE_LIBZ */
    {
      free(buffer); return ERROR_NONE;
    }
}

/* Hardware maps for each version */

/* Version 1 */

hardwareMap z80_version1Map[] = {

  UNSUPPORTED_128KMACHINES,
  { UNSUPPORTED_ANY, PERIPH_IF1, UNSUPPORTED_NOCHANGE, 0, PERIPH_IF1,
    "Warning: %s does not support Interface 1; Interface 1 disabled\n" },
  { UNSUPPORTED_ANY, PERIPH_MGT, UNSUPPORTED_NOCHANGE, 0, PERIPH_MGT,
    "Warning: %s does not support MGT; MGT disabled\n" },
  UNSUPPORTED_END

};

/* Version 2 */

hardwareMap z80_version2Map[] = {

  UNSUPPORTED_NEWMACHINES,
  { MACHINE_128K, PERIPH_SAMRAM, UNSUPPORTED_NOCHANGE, 0, PERIPH_SAMRAM,
    "Warning: %s does not support 128K with SamRam; SamRam disabled\n" },
  { UNSUPPORTED_ANY, PERIPH_MGT, UNSUPPORTED_NOCHANGE, 0, PERIPH_MGT,
    "Warning: %s does not support MGT; MGT disabled\n" },
  { UNSUPPORTED_ANY, PERIPH_IF1 | PERIPH_SAMRAM, UNSUPPORTED_NOCHANGE, 0,
    PERIPH_IF1,
    "Warning: %s does not support Interface I and SamRam\n"
    "simultaneously; Interface I disabled\n" },
  UNSUPPORTED_END

};

/* Version 3 */

hardwareMap z80_version3Map[] = {

  { MACHINE_PLUS2, 0, MACHINE_128K, 0, 0,
    "Warning: %s does not support +2; changing to 128K\n" },
  { MACHINE_PLUS2A, 0, MACHINE_PLUS3, 0, 0,
    "Warning: %s does not support +2A; changing to +3\n" },
  { MACHINE_128K, PERIPH_SAMRAM, UNSUPPORTED_NOCHANGE, 0, PERIPH_SAMRAM,
    "Warning: %s does not support 128K with SamRam; SamRam disabled\n" },
  { MACHINE_PLUS3, PERIPH_SAMRAM, UNSUPPORTED_NOCHANGE, 0, PERIPH_SAMRAM,
    "Warning: %s does not support +3 with SamRam; SamRam disabled\n" },
  { UNSUPPORTED_ANY, PERIPH_MGT | PERIPH_SAMRAM, UNSUPPORTED_NOCHANGE, 0,
    PERIPH_MGT,
    "Warning: %s does not support MGT and SamRam simultaneously;\n"
    "MGT disabled\n" },
  { UNSUPPORTED_ANY, PERIPH_IF1 | PERIPH_SAMRAM, UNSUPPORTED_NOCHANGE, 0,
    PERIPH_IF1,
    "Warning: %s does not support Interface I and SamRam\n"
    "simultaneously; Interface I disabled\n" },
  UNSUPPORTED_END

};
