/* zxs.c: routines for dealing with .zxs format files
   Copyright (c) 1998 Philip Kendall

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: pak21@cam.ac.uk (until at least June 1999)
   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#include "hardware.h"
#include "snapconv.h"
#include "snapinfo.h"

#include "zxs.h"

int zxs_majorVersion=1;
int zxs_minorVersion=2;

typedef int (*zxs_readChunkFunc)(FILE *f,struct snapinfo *sptr,
				 unsigned char **pages,unsigned long length);

typedef struct zxs_readChunkStruct {
  char ID[5];
  zxs_readChunkFunc func;
} zxs_readChunkStruct;

zxs_readChunkStruct zxs_readKnownChunks[] = {
  { "RIFF", zxs_readParentChunk },
  { "LIST", zxs_readParentChunk },
  { "fmtz", zxs_readfmtzChunk },
  { "rZ80", zxs_readrZ80Chunk },
  { "r048", zxs_readr048Chunk },
  { "r128", zxs_readr128Chunk },
  { "r+3 ", zxs_readrPlus3Chunk },
  { "ram0", zxs_readram0Chunk },
  { "ram1", zxs_readram1Chunk },
  { "ram2", zxs_readram2Chunk },
  { "ram3", zxs_readram3Chunk },
  { "ram4", zxs_readram4Chunk },
  { "ram5", zxs_readram5Chunk },
  { "ram6", zxs_readram6Chunk },
  { "ram7", zxs_readram7Chunk }
};

static int zxs_numReadKnownChunks=15;

typedef int (*zxs_writeChunkFunc)(FILE *f,struct snapinfo *sptr,
				  unsigned char **pages,zxs_chunkList *list,
				  unsigned long *length);

typedef struct zxs_writeChunkStruct {
  char ID[5];
  zxs_writeChunkFunc func;
} zxs_writeChunkStruct;

zxs_writeChunkStruct zxs_writeKnownChunks[] = {
  { "RIFF", zxs_writeRIFFChunk },
  { "LIST", zxs_writeLISTChunk },
  { "fmtz", zxs_writefmtzChunk },
  { "rZ80", zxs_writerZ80Chunk },
  { "r048", zxs_writer048Chunk },
  { "r128", zxs_writer128Chunk },
  { "r+3 ", zxs_writerPlus3Chunk },
  { "ram0", zxs_writeram0Chunk },
  { "ram1", zxs_writeram1Chunk },
  { "ram2", zxs_writeram2Chunk },
  { "ram3", zxs_writeram3Chunk },
  { "ram4", zxs_writeram4Chunk },
  { "ram5", zxs_writeram5Chunk },
  { "ram6", zxs_writeram6Chunk },
  { "ram7", zxs_writeram7Chunk }
};

static int zxs_numWriteKnownChunks=15;

int zxs_status=0;

zxs_chunkList *zxs_unknownChunks=0;
zxs_chunkList *zxs_shortChunks  =0;
zxs_chunkList *zxs_longChunks   =0;

static int zxs_readramChunk(FILE *f,struct snapinfo *sptr,unsigned char **page,
			    unsigned long length,const char *ID);
static unsigned long makeDWord(unsigned char *p);
static unsigned long readDWord(FILE *f);

static int zxs_writeParentChunk(FILE *f,struct snapinfo *sptr,
				unsigned char **pages,zxs_chunkList *list,
				unsigned long *length);
static int zxs_writeramChunk(FILE *f,struct snapinfo *sptr,
			     unsigned char *page,unsigned long *length);
static int writeWord(unsigned word,FILE *f);
static int writeDWord(unsigned long d,FILE *f);

static int zxs_inflateBlock(unsigned char **uncomp,
			    unsigned long *uncompLength,unsigned long length,
			    FILE *f,const char *ID);
static int zxs_deflateBlock(unsigned char *uncomp,unsigned long uncompLength,
			    unsigned long *length,FILE *f);

int zxs_readChunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		  unsigned long *bytesread)
{
  unsigned long subChunkLength; char ID[5]; int i,done,error;

  if( (error=zxs_readChunkHeader(f,ID,&subChunkLength)) != 0) return error;
  if(verbose>=1) fprintf(stderr,"Read chunk ID `%s', length %ld\n",
			 ID,subChunkLength);
  (*bytesread)+=8;
  done=0;

  for(i=0; (i<zxs_numReadKnownChunks) && (!done) ;i++)
    {
      if( !strcmp(ID,zxs_readKnownChunks[i].ID) )
	{
	  error=(zxs_readKnownChunks[i].func)(f,sptr,pages,subChunkLength);
	  if(error!=ERROR_NONE) return error;
	  done=1;
	}
    }

  if(!done) /* Unknown chunk ID */
    {
      if(verbose>=2) fprintf(stderr,"Unknown chunk ID: %s\n",ID);

      fseek(f,(long)subChunkLength,SEEK_CUR);
      error=zxs_addChunkToList(&zxs_unknownChunks,ID);
      if(error!=ERROR_NONE) return error;
      zxs_status|=0x01;
    }

  if(subChunkLength%2)
    {
      fseek(f,1,SEEK_CUR);
      (*bytesread)++;
    }

  (*bytesread)+=subChunkLength;

  return ERROR_NONE;
}


int zxs_readChunkHeader(FILE *f,char *ID,unsigned long *length)
{
  unsigned char buffer[4];

  if( fread(ID,4,1,f)!=1 ) return ERROR_READ; ID[4]='\0';
  if( fread(buffer,4,1,f)!=1) return ERROR_READ;
  *length=makeDWord(buffer);

  return ERROR_NONE;

}

int zxs_readParentChunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
			unsigned long length)
{
  unsigned long read; int error;

  fseek(f,4,SEEK_CUR); /* Skip the form field */
  read=4;	/* Read 4 bytes */

  while(read<length)
    {
      if( (error=zxs_readChunk(f,sptr,pages,&read)) != 0) return error;
    }

  return (read==length) ? ERROR_NONE : ERROR_SUBCHUNKLENGTHS;

}

int zxs_readfmtzChunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  unsigned char buffer[8]; int error;

  if(verbose>=2) fprintf(stderr,"Chunk `fmtz': length %ld, expected 8\n",
			 length);

  if(length<8)	/* shorter than expected => ignore */
    {
      fseek(f,(long)length,SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_shortChunks,"fmtz")) != 0 ) return error;
      zxs_status|=0x02;
      return ERROR_NONE;
    }

  if ( fread(buffer,8,1,f)!=1 ) return ERROR_READ;

  sptr->version=buffer[0]+buffer[1]*0x100;

  switch(buffer[2]+buffer[3]*0x100)
    {
    case 0x0010: sptr->machine=MACHINE_48K; sptr->issue2=1; break;
      /* Issue 2 48K */
    case 0x0020: sptr->machine=MACHINE_48K; sptr->issue2=0; break;
      /* Issue 3 48K */
    case 0x0030: sptr->machine=MACHINE_128K;   break; /* 128K */
    case 0x0040: sptr->machine=MACHINE_PLUS2;  break; /* +2 */
    case 0x0050: sptr->machine=MACHINE_PLUS2A; break; /* +2A */
    case 0x0060: sptr->machine=MACHINE_PLUS3;  break; /* +3 */
    }

  /* Ignore the hardware flags bit, as it doesn't contain
     any useful information at the moment
  */

  sptr->compressed=buffer[6]+buffer[7]*0x100;

  if(verbose>=2)
    {
      const char *ptr;

      switch(sptr->machine)
	{
	  case MACHINE_48K   : ptr="48K"; break;
	  case MACHINE_128K  : ptr="128K"; break;
	  case MACHINE_PLUS2 : ptr="+2"; break;
	  case MACHINE_PLUS2A: ptr="+2A"; break;
	  case MACHINE_PLUS3 : ptr="+3"; break;
	  default            : ptr="Internal error: unknown!"; break;
	}

      fprintf(stderr,"File version number: 0x%04x\n",sptr->version);
      fprintf(stderr,"Machine: %s (Issue 2 emulation: %s)\n",ptr,
	      (sptr->issue2 ? "On" : "Off") );
      fprintf(stderr,"Compression method: 0x%04x\n",sptr->compressed);
    }


  if(length>8)
    {
      fseek(f,(long)(length-8),SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_longChunks,"fmtz")) != 0 ) return error;
      zxs_status|=0x04;
    }

  return ERROR_NONE;

}

int zxs_readrZ80Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{unsigned char buffer[33]; int error;

  if(verbose>=2) fprintf(stderr,"Chunk `rZ80': length %ld, expected 33\n",
			 length);

  if(length<33)	/* shorter than expected => ignore */
    {
      fseek(f,(long)length,SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_shortChunks,"rZ80")) != 0 ) return error;
      zxs_status|=0x02;
      return ERROR_NONE;
    }

  if ( fread(buffer,33,1,f)!=1 ) return ERROR_READ;

  sptr->a =buffer[ 0]; sptr->f  =buffer[ 1];
  sptr->bc =buffer[ 2]+buffer[ 3]*0x100;
  sptr->de =buffer[ 4]+buffer[ 5]*0x100;
  sptr->hl =buffer[ 6]+buffer[ 7]*0x100;

  sptr->a_ =buffer[ 8]; sptr->f_=buffer[ 9];
  sptr->bc_=buffer[10]+buffer[11]*0x100;
  sptr->de_=buffer[12]+buffer[13]*0x100;
  sptr->hl_=buffer[14]+buffer[15]*0x100;

  sptr->ix =buffer[16]+buffer[17]*0x100;
  sptr->iy =buffer[18]+buffer[19]*0x100;
  sptr->pc =buffer[20]+buffer[21]*0x100;
  sptr->sp =buffer[22]+buffer[23]*0x100;

  sptr->i  =buffer[24]; sptr->r =buffer[25];

  sptr->iff1=!!buffer[26]; sptr->iff2=!!buffer[27];

  sptr->im=(buffer[28]&0x03);

  sptr->tstates=makeDWord(&buffer[29]);

  if(verbose>=2)
    {
      fprintf(stderr,"AF =0x%02x%02x BC =0x%04x\n",sptr->a,sptr->f,sptr->bc);
      fprintf(stderr,"DE =0x%04x HL =0x%04x\n",sptr->de,sptr->hl);
      fprintf(stderr,"AF'=0x%02x%02x BC'=0x%04x\n",
	      sptr->a_,sptr->f_,sptr->bc_);
      fprintf(stderr,"DE'=0x%04x HL'=0x%04x\n",sptr->de_,sptr->hl_);
      fprintf(stderr,"IX =0x%04x IY =0x%04x\n",sptr->ix,sptr->iy);
      fprintf(stderr,"PC =0x%04x SP =0x%04x\n",sptr->pc,sptr->sp);
      fprintf(stderr," I =0x%02x    R =0x%02x\n",sptr->i,sptr->r);
      fprintf(stderr,"IFF1=%d IFF2=%d IM %d\n",sptr->iff1,sptr->iff1,sptr->im);
      fprintf(stderr,"T states since last interrupt: %ld\n",sptr->tstates);
    }

  if(length>33)
    {
      fseek(f,(long)(length-33),SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_longChunks,"rZ80")) != 0 ) return error;
      zxs_status|=0x04;
    }

  return ERROR_NONE;

}

int zxs_readr048Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  unsigned char buffer[9]; int error;

  if(verbose>=2) fprintf(stderr,"Chunk `r048': length %ld, expected 9\n",
			 length);

  if(length<9)	/* shorter than expected => ignore */
    {
      fseek(f,(long)length,SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_shortChunks,"r048")) != 0 ) return error;
      zxs_status|=0x02;
      return ERROR_NONE;
    }

  if ( fread(buffer,9,1,f)!=1 ) return ERROR_READ;

  sptr->outULA=buffer[0];
  memcpy(sptr->keys,&buffer[1],8);

  if(verbose>=2)
    {
      int i;

      fprintf(stderr,"Last byte output to ULA: 0x%02x\n",sptr->outULA);
      fprintf(stderr,"Key states: ");
      for(i=0;i<7;i++) fprintf(stderr,"0x%02x, ",sptr->keys[i]);
      fprintf(stderr,"0x%02x\n",sptr->keys[7]);
    }

  if(length>9)
    {
      fseek(f,(long)(length-9),SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_longChunks,"r048")) != 0 ) return error;
      zxs_status|=0x04;
    }

  return ERROR_NONE;

}

int zxs_readr128Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  unsigned char buffer[18]; int error;

  if(verbose>=2) fprintf(stderr,"Chunk `r128': length %ld, expected 18\n",
			 length);

  if(length<18)	/* shorter than expected => ignore */
    {
      fseek(f,(long)length,SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_shortChunks,"r128")) != 0 ) return error;
      zxs_status|=0x02;
      return ERROR_NONE;
    }

  if ( fread(buffer,18,1,f)!=1 ) return ERROR_READ;

  sptr->out7ffd=buffer[0]; sptr->outfffd=buffer[1];
  memcpy(sptr->soundchip,&buffer[2],16);

  if(verbose>=2)
    {
      int i;

      fprintf(stderr,"Last OUT to 0x7ffd: 0x%02x\n",sptr->out7ffd);
      fprintf(stderr,"Last OUT to 0xfffd: 0x%02x\n",sptr->outfffd);
      fprintf(stderr,"Soundchip registers: ");
      for(i=0;i<16;i++)
	{
	  fprintf(stderr,"0x%02x",sptr->soundchip[i]);
	  if(i==7) fprintf(stderr,"\n                     ");
	  else if(i==15) fprintf(stderr,"\n");
	  else fprintf(stderr,", ");
	}
    }

  if(length>18)
    {
      fseek(f,(long)(length-18),SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_longChunks,"r128")) != 0 ) return error;
      zxs_status|=0x04;
    }

  return ERROR_NONE;

}

int zxs_readrPlus3Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
			unsigned long length)
{
  unsigned char buffer[1]; int error;

  if(verbose>=2) fprintf(stderr,"Chunk `r+3 ': length %ld, expected 1\n",
			 length);

  if(length<1)	/* shorter than expected => ignore */
    {
      fseek(f,(long)(length),SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_shortChunks,"r+3 ")) != 0 ) return error;
      zxs_status|=0x02;
      return ERROR_NONE;
    }

  if ( fread(buffer,1,1,f)!=1 ) return ERROR_READ;

  sptr->out1ffd=buffer[0];

  if(verbose>=2) fprintf(stderr,"Last OUT to 0x1ffd: 0x%02x\n",sptr->out1ffd);

  if(length>1)
    {
      fseek(f,(long)(length-1),SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_longChunks,"r+3 ")) != 0 ) return error;
      zxs_status|=0x04;
      return ERROR_NONE;
    }

  return ERROR_NONE;

}

int zxs_readram0Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[0],length,"ram0");
}

int zxs_readram1Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[1],length,"ram1");
}

int zxs_readram2Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[2],length,"ram2");
}

int zxs_readram3Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[3],length,"ram3");
}

int zxs_readram4Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[4],length,"ram4");
}

int zxs_readram5Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[5],length,"ram5");
}

int zxs_readram6Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[6],length,"ram6");
}

int zxs_readram7Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		      unsigned long length)
{
  return zxs_readramChunk(f,sptr,&pages[7],length,"ram7");
}

static int zxs_readramChunk(FILE *f,struct snapinfo *sptr,unsigned char **page,
			    unsigned long length,const char *ID)
{
  unsigned long pageLength;
  int error;

  switch(sptr->compressed)
    {
      case 0x0008: /* Deflated */

	error=zxs_inflateBlock(page,&pageLength,length,f,ID);
	if(error!=ERROR_NONE) return error;

	/* If the page has not decompressed to 16384 bytes, this is an
           error; pageLength=0 => compression header too short and the
           chunk was skipped */
	if(pageLength!=0 && pageLength!=0x4000)
	  {
	    free(*page); (*page)=0;
	    return ERROR_PAGELENGTH;
	  }

	break;

      case 0xffff: /* Not compressed */

	if( length!=0x4000 ) return ERROR_PAGELENGTH;
	if( ((*page)=(unsigned char*)malloc(0x4000*sizeof(unsigned char)))==0)
	  return ERROR_MEMORY;
	if( fread((*page),0x4000,1,f)!=1 )
	  {
	    free(*page); (*page)=0;
	    return ERROR_READ;
	  }

	break;

      default: /* Anything else */
	return ERROR_COMPRESSED; /* Can't handle */
    }

  return ERROR_NONE;

}

static unsigned long makeDWord(unsigned char *p)
{
  return *p + *(p+1)*0x100 + *(p+2)*0x10000 + *(p+3)*0x1000000;
}

static unsigned long readDWord(FILE *f)
{
  return fgetc(f)+fgetc(f)*0x100+fgetc(f)*0x10000+fgetc(f)*0x1000000;
}

/* Write functions start here */

int zxs_writeChunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		   zxs_chunkList *list,unsigned long *length)
{
  int i,done=0,error;

  if(verbose>1) fprintf(stderr,"Writing chunk ID `%s'\n",list->ID);

  for(i=0; (i<zxs_numWriteKnownChunks) && (!done); i++ )
    {
      if(!strcmp(list->ID,zxs_writeKnownChunks[i].ID))
	{
	  error=(zxs_writeKnownChunks[i].func)(f,sptr,pages,list->sublist,
					       length);
	  if(error!=ERROR_NONE) return error;

	  done=1;
	}
    }

  if(!done) return ERROR_UNKNOWNCHUNK;

  return 0;

}

int zxs_writeRIFFChunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{
  fputs("RIFF",f); /* ID */
  (*length)+=4;

  return zxs_writeParentChunk(f,sptr,pages,list,length);
}

int zxs_writeLISTChunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{
  fputs("LIST",f); /* ID */
  (*length)+=4;

  return zxs_writeParentChunk(f,sptr,pages,list,length);
}

static int zxs_writeParentChunk(FILE *f,struct snapinfo *sptr,
				unsigned char **pages,zxs_chunkList *list,
				unsigned long *length)
{

  int error;

  unsigned long chunkLength,  /* How long is this chunk */
                lengthOffset; /* Where the length will be stored */

  if(list==0) return ERROR_NOFORMTYPE;

  lengthOffset=ftell(f);
  fwrite("\0\0\0\0",4,1,f); /* Length; will be overwritten with the correct
			       value when we know it :-) */
  (*length)+=4;

  /* Write out the form of this parent chunk */
  fwrite(list->ID,4,1,f); chunkLength=4;

  /* And move onto the first real chunk */
  list=list->next;

  while(list)
    {
      error=zxs_writeChunk(f,sptr,pages,list,&chunkLength);
      if(error!=ERROR_NONE) return error;
      list=list->next;
    }

  fseek(f,(long)lengthOffset,SEEK_SET); /* Write the length in */
  writeDWord(chunkLength,f);

  fseek(f,0,SEEK_END); /* Move back to the end of the file */

  if(chunkLength%2)	/* Chunks may not have an odd length */
    {
      fputc('\0',f);
      (*length)++;
    }

  (*length)+=chunkLength; /* And update the length of the parent of this
			     parent chunk */

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}


int zxs_writefmtzChunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{
  fputs("fmtz",f); /* ID */
  writeDWord(8,f); /* Length */

  writeWord(0x0100,f); /* All files created at the moment are v1.00 */

  switch(sptr->machine)
    {
    case 0:  writeWord( (sptr->issue2) ? 0x0010U : 0x0020U , f ); break;
    case 1:  writeWord( 0x0030U,f ); break;
    case 2:  writeWord( 0x0040U,f ); break;
    case 3:  writeWord( 0x0050U,f ); break;
    case 4:  writeWord( 0x0060U,f ); break;

      /* Map unknown machines to standard 128K Spectrum */

    default: writeWord( 0x0030U,f ); break;
    }

  writeWord(0,f); /* No specialised hardware */

#ifdef HAVE_LIBZ
  writeWord(0x0008,f); /* Using deflation as my compression method */
#else /* #ifdef HAVE_LIBZ */
  writeWord(0xffff,f); /* Not compressed */
#endif /* #ifdef HAVE_LIBZ */

  (*length)+=(8+8);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int zxs_writerZ80Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
			zxs_chunkList *list,unsigned long *length)
{

  fputs("rZ80",f); /* ID */
  writeDWord(33,f); /* Length */

  fputc(sptr->a,f); fputc(sptr->f,f);
  writeWord(sptr->bc,f); writeWord(sptr->de,f);
  writeWord(sptr->hl,f);

  fputc(sptr->a_,f); fputc(sptr->f_,f);
  writeWord(sptr->bc_,f); writeWord(sptr->de_,f);
  writeWord(sptr->hl_,f);

  writeWord(sptr->ix,f); writeWord(sptr->iy,f);
  writeWord(sptr->pc,f); writeWord(sptr->sp,f);

  fputc(sptr->i,f); fputc(sptr->r,f);
  fputc(sptr->iff1,f); fputc(sptr->iff2,f);
  fputc(sptr->im,f);

  writeDWord(sptr->tstates,f);

  fputc('\0',f); /* Chunks may not have an odd length */

  (*length)+=(8+34);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int zxs_writer048Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("r048",f); /* ID */
  writeDWord(9,f); /* Length */

  fputc(sptr->outULA,f);
  fwrite(sptr->keys,8,1,f);

  fputc('\0',f); /* Chunks may not have an odd length */

  (*length)+=(8+10);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int zxs_writer128Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("r128",f); /* ID */
  writeDWord(18,f); /* Length */

  fputc(sptr->out7ffd,f); fputc(sptr->outfffd,f);
  fwrite(sptr->soundchip,16,1,f);

  (*length)+=(8+18);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int zxs_writerPlus3Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
			 zxs_chunkList *list,unsigned long *length)
{

  fputs("r+3 ",f); /* ID */
  writeDWord(1,f); /* Length */

  fputc(sptr->out1ffd,f);

  fputc('\0',f); /* Chunks may not have an odd length */

  (*length)+=(8+1);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

int zxs_writeram0Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram0",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[0],length);

}

int zxs_writeram1Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram1",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[1],length);

}

int zxs_writeram2Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram2",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[2],length);

}

int zxs_writeram3Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram3",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[3],length);

}

int zxs_writeram4Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram4",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[4],length);

}

int zxs_writeram5Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram5",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[5],length);

}

int zxs_writeram6Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram6",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[6],length);

}

int zxs_writeram7Chunk(FILE *f,struct snapinfo *sptr,unsigned char **pages,
		       zxs_chunkList *list,unsigned long *length)
{

  fputs("ram7",f); /* ID */
  (*length)+=4;

  return zxs_writeramChunk(f,sptr,pages[7],length);

}

static int zxs_writeramChunk(FILE *f,struct snapinfo *sptr,unsigned char *page,
			     unsigned long *length)
{

  int error; unsigned char *buffer;

  if(page) buffer=page;
  else
    {
      if( (buffer=malloc(0x4000*sizeof(unsigned char)))==NULL )
	return ERROR_MEMORY;
      memset(buffer,0,0x4000);
    }

  if(ferror(f)) { return 1; }

#ifdef HAVE_LIBZ

  error=zxs_deflateBlock(buffer,0x4000,length,f);

#else /* #ifdef HAVE_LIBZ */

  writeDWord(0x4000,f); /* Length */

  fwrite(buffer,0x4000,1,f); /* Data */

  (*length)+=4+0x4000;

  error= ferror(f) ? ERROR_WRITE : ERROR_NONE;

#endif /* #ifdef HAVE_LIBZ */

  if(buffer!=page) free(buffer);

  if(error!=ERROR_NONE) return error;

  return ERROR_NONE;

}

static int writeWord(unsigned w,FILE *f)
{

  fputc((int)(w%0x100),f); fputc((int)(w/0x100),f);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

static int writeDWord(unsigned long d,FILE *f)
{

  fputc((int)(  d           %0x100 ), f);
  fputc((int)( (d/    0x100)%0x100 ), f);
  fputc((int)( (d/  0x10000)%0x100 ), f);
  fputc((int)( (d/0x1000000)%0x100 ), f);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

#ifdef HAVE_LIBZ

static int zxs_inflateBlock(unsigned char **uncomp,
			    unsigned long *uncompLength,unsigned long length,
			    FILE *f,const char *ID)
{
  unsigned char *buffer; int error;
  unsigned long crc1,crc2,expectedLength,headerLength;

  headerLength=readDWord(f); if(ferror(f)) return ERROR_READ;

  if(verbose>=2)
    fprintf(stderr,"Compression info header has length %ld, expected 12\n",
	    headerLength);

  /* If the compression information header is smaller than expected,
     ignore this chunk */
  if(headerLength<12)
    {
      fseek(f,length-4,SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_shortChunks,ID)) != 0 )
	return error;
      zxs_status|=0x02;
      (*uncompLength)=0;
      return ERROR_NONE;
    }

  buffer=malloc(8*sizeof(unsigned char));
  if(buffer==0) return ERROR_MEMORY;

  if( fread(buffer,8,1,f)==-1 ) { free(buffer); return ERROR_READ; }

  crc1=makeDWord(buffer);
  expectedLength=makeDWord(&buffer[4]);

  /* If the compression info header is longer than expected,
     skip the unknown bits and carry on */
  if(headerLength>12)
    {
      fseek(f,headerLength-12,SEEK_CUR);
      if( (error=zxs_addChunkToList(&zxs_longChunks,ID)) != 0 )
	{ free(buffer); return error; }
      zxs_status|=0x04;
    }

  /* Allocate some space for the compressed data; +6 is to allow
     for zlib's header/footer */
  buffer=realloc(buffer,length-headerLength+6);
  if(buffer==0) return ERROR_MEMORY;

  /* Read the compressed data in */
  if( fread(&buffer[2],length-headerLength,1,f)==-1 )
    { free(buffer); return ERROR_READ; }

  /* Put the zlib header back on */
  buffer[0]=0x78; buffer[1]=0xda;

  /* Allocate some space to store the uncompressed data in */
  if( ((*uncomp)=(unsigned char*)malloc(0x4000*sizeof(unsigned char)))==0)
    { free(buffer); return ERROR_MEMORY; }
  (*uncompLength)=0x4000;

  /* And uncompress the data */
  error=uncompress(*uncomp,uncompLength,buffer,length-headerLength+6);

  /* At this point, we _expect_ to have got a Z_DATA_ERROR, as the
     Adler-32 checksum (as used by zlib) isn't stored in the .zxs
     file. This is a horrible kludge, I know :-( */
  if( (error!=Z_OK && error!=Z_DATA_ERROR) )
    { free(buffer); free(*uncomp); (*uncomp)=0; return ERROR_COMPRESSION; }

  /* Check the data expanded to the expected length */
  if( *uncompLength != expectedLength )
    { free(buffer); free(*uncomp); (*uncomp)=0; return ERROR_CORRUPT; }

  /* Don't need buffer anymore */
  free(buffer);

  /* We can however, check the CRC */
  crc2=crc32(0L,Z_NULL,0);
  crc2=crc32(crc2,*uncomp,*uncompLength);
  if(crc1!=crc2) { free(*uncomp); (*uncomp)=0; return ERROR_CORRUPT; }

  return ERROR_NONE;

}

static int zxs_deflateBlock(unsigned char *uncomp,unsigned long uncompLength,
			    unsigned long *length,FILE *f)
{
  unsigned char *buffer; unsigned long compLength=uncompLength+128,error,crc;

  buffer=(unsigned char*)malloc(compLength*sizeof(unsigned char));
  if(buffer==0) return ERROR_MEMORY;

  /* Compress the data */
  error=compress2(buffer,&compLength,uncomp,uncompLength,Z_BEST_COMPRESSION);
  if(error!=Z_OK) { free(buffer); return ERROR_COMPRESSION; }

  /* Work out the CRC of the input data */
  crc=crc32(0L,Z_NULL,0);
  crc=crc32(crc,uncomp,uncompLength);

  /* Write out the length, minus 6 bytes for the zlib header/footer, plus
     12 bytes for zx32's header */
  writeDWord(compLength-6+12,f);
  (*length)+=4;

  writeDWord(12,f); /* Size of zx32's header */
  writeDWord(crc,f);
  writeDWord(uncompLength,f); /* Uncompressed size */
  (*length)+=12;

  fwrite(&buffer[2],compLength-6,1,f);
  (*length)+=compLength-6;

  /* Chunks may not have an odd length */
  if( (compLength-6+12)%2 ) { fputc('\0',f); (*length)++; }

  free(buffer);

  return ferror(f) ? ERROR_WRITE : ERROR_NONE;

}

#else /* #ifdef HAVE_LIBZ */

static int zxs_inflateBlock(unsigned char **uncomp,
			    unsigned long *uncompLength,unsigned long length,
			    FILE *f,const char *ID)
{
  return ERROR_COMPRESSED;
}

static int zxs_deflateBlock(unsigned char *uncomp,unsigned long uncompLength,
			    unsigned long *length,FILE *f)
{
  return ERROR_COMPRESSED;
}

#endif /* #ifdef HAVE_LIBZ */

/* Utility functions */

int zxs_addChunkToList(zxs_chunkList **list,const char *ID)
{

  zxs_chunkList *p;

  if((*list)==0)
    {
      (*list)=(zxs_chunkList*)malloc(sizeof(zxs_chunkList));
      if((*list)==0) return ERROR_MEMORY;
      p=(*list);
    }
  else
    {
      p=(*list); while(p->next) p=p->next;
      p->next=(zxs_chunkList*)malloc(sizeof(zxs_chunkList));
      if( (p->next)==0 ) return ERROR_MEMORY;
      p=p->next;
    }

  memcpy(p->ID,ID,4); (p->ID)[4]='\0';
  p->sublist=p->next=0;

  return ERROR_NONE;

}

void zxs_cleanUp()
{
  zxs_deleteList(&zxs_unknownChunks);
  zxs_deleteList(&zxs_shortChunks  );
  zxs_deleteList(&zxs_longChunks   );
}

void zxs_deleteList(zxs_chunkList **list)
{
  zxs_chunkList *p1,*p2;

  p2=(*list);

  if(p2!=0)
    {
      p1=p2->next;

      while(p1)
	{
	  if(p2->sublist) zxs_deleteList(&(p2->sublist)); free(p2);
	  p2=p1; p1=p1->next;
	}

      if(p2->sublist) zxs_deleteList(&(p2->sublist)); free(p2);

    }

  (*list)=0;

}

/* The hardware map */

hardwareMap zxs_hardwareMap[] = {

  UNSUPPORTED_PERIPHS,
  UNSUPPORTED_END

};
